# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

from typing import List, Dict

from .csv_logging import CSVLogger
from .common_stream import CommonStream
from ..core.enums.dcb_enums import DCBCommand
from ..core.enums.ad7156_enums import AD7156Command
from ..core.packets.command_packet import CommandPacket
from ..core.packets.stream_data_packets import AD7156DataPacket
from ..core.enums.common_enums import Application, CommonCommand, Stream
from ..core.packets.ad7156_packets import AD7156DCBPacket, AD7156RegisterReadPacket, AD7156RegisterWritePacket, \
    AD7156DCBCommandPacket


class AD7156Application(CommonStream):
    """
    AD7156 Application class.

    .. code-block:: python3
        :emphasize-lines: 4

        from adi_study_watch import SDK

        sdk = SDK("COM4")
        application = sdk.get_ad7156_application()
    """

    def __init__(self, packet_manager):
        super().__init__(Application.AD7156, Stream.AD7156, packet_manager)
        self._dcb_size = 20

    @staticmethod
    def get_supported_streams() -> List[Stream]:
        """
        List all supported streams.

        :return: Array of stream ID enums.
        :rtype: List[Stream]

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.get_supported_streams()
            print(x)
            # [<Stream.ADPD1: ['0xC2', '0x11']>, ... , <Stream.ADPD12: ['0xC2', '0x1D']>]
        """
        return [Stream.AD7156]

    def _callback_data(self, packet, packet_id, callback_function=None, args=None):
        """
        Process and returns the data back to user's callback function.
        """
        self._callback_data_helper(packet, AD7156DataPacket())

    def delete_device_configuration_block(self) -> Dict:
        """
        Deletes AD7156 Device configuration block.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ad7156_application()
            application.delete_device_configuration_block()
        """
        request_packet = AD7156DCBCommandPacket(self._destination, DCBCommand.ERASE_CONFIG_REQ)
        response_packet = AD7156DCBCommandPacket(self._destination, DCBCommand.ERASE_CONFIG_RES)
        return self._send_packet(request_packet, response_packet)

    def load_configuration(self) -> Dict:
        """
        Loads configuration.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ad7156_application()
            x = application.load_configuration()
            print(x["payload"]["status"])
            # CommonStatus.OK
        """
        request_packet = CommandPacket(self._destination, AD7156Command.LOAD_CONFIG_REQ)
        response_packet = CommandPacket(self._destination, AD7156Command.LOAD_CONFIG_RES)
        return self._send_packet(request_packet, response_packet)

    def read_device_configuration_block(self) -> Dict:
        """
        Returns entire device configuration block.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ad7156_application()
            x = application.read_device_configuration_block()
            print(x["payload"]["data"])
            # []
        """
        request_packet = AD7156DCBCommandPacket(self._destination, DCBCommand.READ_CONFIG_REQ)
        response_packet = AD7156DCBPacket(self._destination, DCBCommand.READ_CONFIG_RES)
        return self._send_packet(request_packet, response_packet)

    def read_register(self, addresses: List[int]) -> Dict:
        """
        Reads the register values of specified addresses. This function takes a list of addresses to read,
        and returns a response packet as dictionary containing addresses and values.

        :param addresses: List of register addresses to read.
        :type addresses: List[int]
        :return: A response packet as dictionary
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x00
             - 0x17

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ad7156_application()
            x = application.read_register([0x10, 0x12])
            print(x["payload"]["data"])
            # [['0x10', '0x0'], ['0x12', '0x0']]
        """
        data = [[address, 0] for address in addresses]
        request_packet = AD7156RegisterReadPacket(self._destination, CommonCommand.REGISTER_READ_REQ)
        request_packet.set_payload("size", len(data))
        request_packet.set_payload("data", data)
        response_packet = AD7156RegisterReadPacket(self._destination, CommonCommand.REGISTER_READ_RES)
        return self._send_packet(request_packet, response_packet)

    def write_device_configuration_block(self, addresses_values: List[List[int]]) -> Dict:
        """
        Writes the device configuration block values of specified addresses.
        This function takes a list of addresses and values to write, and returns a response packet as
        dictionary containing addresses and values.

        :param addresses_values: List of addresses and values to write.
        :type addresses_values: List[List[int]]
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x09
             - 0x12

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ad7156_application()
            x = application.write_device_configuration_block([[0x10, 2], [0x12, 0x1]])
            print(x["payload"]["size"])
            # 2
        """
        request_packet = AD7156DCBPacket(self._destination, DCBCommand.WRITE_CONFIG_REQ)
        request_packet.set_payload("size", len(addresses_values))
        request_packet.set_payload("data", addresses_values)
        response_packet = AD7156DCBCommandPacket(self._destination, DCBCommand.WRITE_CONFIG_RES)
        return self._send_packet(request_packet, response_packet)

    def write_device_configuration_block_from_file(self, filename: str) -> Dict:
        """
        Writes the device configuration block values of specified addresses from file.

        :param filename: dcb filename
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x09
             - 0x12

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ad7156_application()
            application.write_device_configuration_block_from_file("adxl_dcb.dcfg")
        """
        result = self.device_configuration_file_to_list(filename)
        if result:
            return self.write_device_configuration_block(result)

    def write_register(self, addresses_values: List[List[int]]) -> Dict:
        """
        Writes the register values of specified addresses. This function takes a list of addresses and values to write,
        and returns a response packet as dictionary containing addresses and values.

        :param addresses_values: List of register addresses and values to write.
        :type addresses_values: List[List[int]]
        :return: A response packet as dictionary.
        :rtype: Dict

        .. list-table::
           :widths: 50 50
           :header-rows: 1

           * - Address Lower Limit
             - Address Upper Limit
           * - 0x09
             - 0x12

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ad7156_application()
            x = application.write_register([[0x10, 0x1], [0x12, 0x2]])
            print(x["payload"]["data"])
            # [['0x10', '0x1'], ['0x12', '0x2']]
        """
        request_packet = AD7156RegisterWritePacket(self._destination, CommonCommand.REGISTER_WRITE_REQ)
        request_packet.set_payload("size", len(addresses_values))
        request_packet.set_payload("data", addresses_values)
        response_packet = AD7156RegisterWritePacket(self._destination, CommonCommand.REGISTER_WRITE_RES)
        return self._send_packet(request_packet, response_packet)

    def enable_csv_logging(self, filename: str, header: List = None) -> None:
        """
        Start logging stream data into CSV.

        :param filename: Name of the CSV file.
        :param header: Header list of the CSV file.
        :return: None

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ad7156_application()
            x = application.enable_csv_logging("ad7156.csv")
        """
        if header is None:
            header = ["Timestamp", "CH1", "CH2", "CH1 ADCCode", "CH2 ADCCode", "OUT1 val", "OUT2 val"]
        self._csv_logger[Stream.AD7156] = CSVLogger(filename, header)

    def disable_csv_logging(self) -> None:
        """
        Stops logging stream data into CSV.

        :return: None

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_ad7156_application()
            x = application.disable_csv_logging()
        """
        if self._csv_logger[Stream.AD7156]:
            self._csv_logger[Stream.AD7156].stop_logging()
        self._csv_logger[Stream.AD7156] = None



# ******************************************************************************
# Copyright (c) 2023 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

"""
    Interprocess communication tools
"""
import datetime
import queue
import sys
import threading
import time
from typing import Optional, Any, Iterable, List


class BaseThread(threading.Thread):
    """
        Extends Thread functionality by adding Event to gracefully shut down the thread
        and start delay.
    """

    TIMESTAMP_TO_MILLISECONDS = 1e3

    def __init__(self):
        threading.Thread.__init__(self, target=self._run)
        self._is_running: threading.Event = threading.Event()
        self.start_delay: int = 0

    def _run(self):

        if self.start_delay > 0:
            time.sleep(self.start_delay)
        while self._is_running.is_set():
            self.loop()

    def start(self, start_delay: int = 0):
        self._is_running.set()
        self.start_delay = start_delay
        super().start()

    def stop(self):
        self._is_running.clear()

    def loop(self):
        raise NotImplemented()

    def is_running(self) -> bool:
        return self._is_running.is_set()

    @staticmethod
    def get_ts(utc: bool = True) -> int:

        if utc:
            dt = datetime.datetime.now(datetime.timezone.utc)
            utc_time = dt.replace(tzinfo=datetime.timezone.utc)
            return int(utc_time.timestamp() * BaseThread.TIMESTAMP_TO_MILLISECONDS)

        return int(time.time() * BaseThread.TIMESTAMP_TO_MILLISECONDS)


class FifoQueue:
    """
        Thread safe FIFO queue implementation

    """

    def __init__(self, maxsize: int = 0):

        self.max_size: int = maxsize
        self._internal_queue: queue.Queue = queue.Queue(maxsize)
        self._q_lock: threading.Lock = threading.Lock()

    def push(self, item: Any) -> Optional[Any]:
        """
            Push one item to FIFO queue
            return: removed older items list if queue is full
        """
        _e: Optional[Any] = None
        with self._q_lock:
            if self._internal_queue.full():
                _e = self._internal_queue.get()
            self._internal_queue.put(item)
        return _e

    def push_all(self, item: Iterable) -> List[Any]:
        """
            Push a list of items to queue as separated.
            If internal queue is full, remove the oldest item.
            return: a list of removed items if any
        """
        _removed: List[Any] = []
        with self._q_lock:
            for _i in item:
                if self._internal_queue.full():
                    _removed.append(self._internal_queue.get())
                self._internal_queue.put(_i)
        return _removed

    def pop(self, timeout: int = 0) -> Optional[Any]:
        """
            Pop one item from FIFO. Will return immediately if there is an item in the queue,
            block and wait with timeout until an item is available in the queue.

            Return none if timeout or no items in the queue.
        """
        with self._q_lock:
            if not self._internal_queue.empty():
                return self._internal_queue.get(block=timeout > 0, timeout=timeout)
        return None

    def pop_all(self, timeout: int = 0, count: int = sys.maxsize, wait_all: bool = False) -> Optional[List[Any]]:
        """
            Pop desired count of items form queue, with timeout or block until all are available.
        """

        result: Optional[List[Any]] = []
        start_time: int = int(time.time())

        # non blocking
        if timeout == 0:
            with self._q_lock:
                if wait_all and self._internal_queue.qsize() >= count:
                    while len(result) < count:
                        result.append(self._internal_queue.get())
                    return result

                if not wait_all:
                    while not self._internal_queue.empty() and len(result) < count:
                        result.append(self._internal_queue.get())
                    return result

            if wait_all:
                return None

        # blocking with timeout
        while int(time.time()) - start_time < timeout:
            with self._q_lock:

                if wait_all and self._internal_queue.qsize() >= count:
                    while len(result) < count:
                        result.append(self._internal_queue.get())
                    return result

                if not wait_all:
                    while not self._internal_queue.empty() and len(result) < count:
                        result.append(self._internal_queue.get())
            time.sleep(1)

        return result if len(result) > 0 else None

    def clear(self) -> None:
        """
            Remove all items from queue
        """
        with self._q_lock:
            self._internal_queue = queue.Queue(self.max_size)

    def empty(self) -> bool:
        """
            Check if queue is empty
        """
        with self._q_lock:
            return self._internal_queue.empty()



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

from enum import Enum

from .. import utils


class BIACommand(Enum):
    """
    BIACommand Enum
    """
    LOWEST = [0x40]
    SET_DFT_NUM_REQ = [0x46]
    SET_DFT_NUM_RES = [0x47]
    SET_HS_TRANS_IMPEDANCE_AMPLIFIER_CAL_REQ = [0x48]
    SET_HS_TRANS_IMPEDANCE_AMPLIFIER_CAL_RES = [0x49]
    FDS_STATUS_REQ = [0x4A]
    FDS_STATUS_RES = [0x4B]
    DCB_TIMING_INFO_REQ = [0x4C]
    DCB_TIMING_INFO_RES = [0x4D]
    ALGO_STREAM_RES = [0x4E]
    LOAD_DCFG_REQ = [0x4F]
    LOAD_DCFG_RES = [0x50]
    WRITE_DCFG_REQ = [0x51]
    WRITE_DCFG_RES = [0x52]
    READ_DCFG_REQ = [0x53]
    READ_DCFG_RES = [0x54]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


class HSResistorTIA(Enum):
    """
    HSResistorTIA Enum
    """
    RESISTOR_200 = [0x0, 0x0]
    RESISTOR_1K = [0x0, 0x1]
    RESISTOR_5K = [0x0, 0x2]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


class BIADFTWindow(Enum):
    """
    BIADFTWindow Enum
    """
    DFT_WINDOW_4 = [0x0, 0x0]
    DFT_WINDOW_8 = [0x0, 0x1]
    DFT_WINDOW_16 = [0x0, 0x2]
    DFT_WINDOW_32 = [0x0, 0x3]
    DFT_WINDOW_64 = [0x0, 0x4]
    DFT_WINDOW_128 = [0x0, 0x5]
    DFT_WINDOW_256 = [0x0, 0x6]
    DFT_WINDOW_512 = [0x0, 0x7]
    DFT_WINDOW_1024 = [0x0, 0x8]
    DFT_WINDOW_2048 = [0x0, 0x9]
    DFT_WINDOW_4096 = [0x0, 0x10]
    DFT_WINDOW_8192 = [0x0, 0x11]
    DFT_WINDOW_16384 = [0x0, 0x12]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


class BIASweepFrequency(Enum):
    """
    BIASweepFrequency Enum
    """
    FREQ_1000HZ = [0x0]
    FREQ_3760HZ = [0x1]
    FREQ_14140HZ = [0x2]
    FREQ_53180HZ = [0x3]
    FREQ_200KHZ = [0x4]
    FREQ_50KHZ = [0xFF]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


class BIAAppInfo(Enum):
    """
    BIAAppInfo Enum
    """
    BITSET_LEADS_OFF = [0x0]
    BITSET_LEADS_ON = [0x1]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))



# ******************************************************************************
# Copyright (c) 2023 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************
"""
    Serial port implementation of the ConnectionInterface
"""
import serial
from .connection_interface import ConnectionInterface


class SerialConnection(ConnectionInterface):
    """
        Serial port access wrapper class. No additional functionality is added.
    """

    def __init__(self, *args, **kwargs):
        ConnectionInterface.__init__(self, *args, **kwargs)
        self.serial: serial.Serial = serial.Serial(*args, **kwargs)

    def read(self, size=1) -> bytes:
        return self.serial.read(size)

    def write(self, data) -> int:
        return self.serial.write(data)

    def close(self) -> None:
        self.serial.close()

    def open(self) -> None:
        self.serial.open()

    def cancel_read(self) -> None:
        self.serial.cancel_read()

    def is_connected(self) -> bool:
        return self.serial.isOpen()




# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

from enum import Enum

from .. import utils


class DCBCommand(Enum):
    """
    DCBCommand Enum
    """
    READ_CONFIG_REQ = [0x97]
    READ_CONFIG_RES = [0x98]
    WRITE_CONFIG_REQ = [0x99]
    WRITE_CONFIG_RES = [0x9A]
    ERASE_CONFIG_REQ = [0x9B]
    ERASE_CONFIG_RES = [0x9C]
    QUERY_STATUS_REQ = [0x9D]
    QUERY_STATUS_RES = [0x9E]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


class DCBStatus(Enum):
    """
    DCBStatus Enum
    """
    TMP_ENUM_FIX = [0x0]
    OK = [0x97]
    ERROR_NO_DCB = [0x98]
    ERROR_NOT_CHKD = [0xFF]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


class DCBConfigBlockIndex(Enum):
    """
    DCBConfigBlockIndex Enum
    """
    GENERAL_BLOCK = [0x0]
    AD5940_BLOCK = [0x1]
    ADPD4000_BLOCK = [0x2]
    ADXL362_BLOCK = [0x3]
    PPG_BLOCK = [0x4]
    ECG_BLOCK = [0x5]
    EDA_LCFG_BLOCK = [0x6]
    AD7156_BLOCK = [0x7]
    PEDOMETER_BLOCK = [0x8]
    TEMPERATURE_BLOCK = [0x9]
    LT_APP_LCFG_BLOCK = [0xA]
    UI_CONFIG_BLOCK = [0xB]
    USER0_BLOCK = [0xC]
    USER1_BLOCK = [0xD]
    USER2_BLOCK = [0xE]
    USER3_BLOCK = [0xF]
    BIA_LCFG_BLOCK = [0x10]
    BIA_DCFG_BLOCK = [0x11]
    EDA_DCFG_BLOCK = [0x12]
    ADP5360_BLOCK = [0x13]
    SESSION_CONFIG_BLOCK = [0x14]
    TEMPERATURE_CORRECTION_BLOCK = [0x15]
    MAX_BLOCK = [0x16]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))



# ******************************************************************************
# Copyright (c) 2023 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************
import os
import time
import logging
from typing import List
from datetime import datetime, timezone, timedelta

from tqdm import tqdm
from serial.tools import list_ports

from .connection import NrfUartBleConnection
from .core import utils
from .core.ble_manager import BLEManager
from .core.enums.board_enums import Board
from .core.enums.fs_enums import FSCommand
from .core.enums.pm_enums import PMCommand
from .core.enums.sqi_enum import SQICommand
from .core.enums.ppg_enums import PPGCommand
from .core.enums.ecg_enums import ECGCommand
from .core.packet_manager import PacketManager
from .application.pm_application import PMApplication
from .core.packets.adpd_packets import ADPDDCFGPacket
from .application.fs_application import FSApplication
from .core.packets.adxl_packets import ADXLDCFGPacket
from .application.eda_application import EDAApplication
from .application.sqi_application import SQIApplication
from .application.ecg_application import ECGApplication
from .application.bia_application import BIAApplication
from .application.ppg_application import PPGApplication
from .core.enums.pedometer_enums import PedometerCommand
from .application.adxl_application import ADXLApplication
from .application.adpd_application import ADPDApplication
from .application.test_application import TestApplication
from .application.user0_application import User0Application
from .application.session_manager_application import SessionManagerApplication
from .core.packets.ecg_packets import ECGLibraryConfigPacket
from .core.packets.ppg_packets import LibraryConfigDataPacket
from .application.ad7156_application import AD7156Application
from .application.adp5360_application import ADP5360Application
from .core.packets.fs_packets import KeyValuePairResponsePacket
from .application.lt_mode4_application import LTMode4Application
from .application.vsm_mb_sb.pm_application import VSMPMApplication
from .application.low_touch_application import LowTouchApplication
from .core.packets.common_packets import VersionPacket, AlarmPacket
from .application.pedometer_application import PedometerApplication
from .core.packets.pm_packets import DateTimePacket, SystemInfoPacket
from .application.temperature_application import TemperatureApplication
from .core.enums.common_enums import Application, CommonCommand, CommonStatus
from .connection.connection_interface import CommunicationMode
from .connection.serial_connection import SerialConnection

logger = logging.getLogger(__name__)


class ConnectionAlreadyExistError(Exception):
    pass


class DongleNotFoundError(Exception):
    pass


class ErroneousParametersCombination(Exception):
    """
        Parameters communication_mod and mac_address are not defined correctly for * CommunicationMode.BleNative
            * CommunicationMode.BleNative - BleNative only. Parameters mac_address and ble_timeout  will be used.
                        Where mac_address is the MAC address of the target device
    """
    pass


class SDK:
    """
    SDK class
    """

    READING_LOG = 0
    JOINING_CSV = 1

    STUDY_WATCH = Board.STUDY_WATCH
    VSM_MB_SB = Board.VSM_MB_SB
    __version__ = "5.22.1"

    def __init__(self, serial_port_address: str, mac_address: str = None, baud_rate: int = 921600,
                 board=Board.STUDY_WATCH, logging_filename: str = None, debug: bool = False,
                 sync_date_time=True, check_version=True, ble_vendor_id: int = 0x0456, ble_product_id: int = 0x2CFE,
                 ble_serial_number: str = None, ble_timeout: int = 10, check_existing_connection=True,
                 communication_mode: CommunicationMode = CommunicationMode.SerialOrDongle,
                 **kwargs):
        """
        Creates a SDK object

        :param serial_port_address: serial port of the device connected.
        :param mac_address: MAC address of the device.
        :param baud_rate: baud rate.
        :param board: board to connect (STUDY_WATCH, VSM_MB_SB).
        :param logging_filename: log file name.
        :param debug: control for debug mode.
        :param sync_date_time: Sync current system date and time with firmware (Default=True).
        :param check_version: Check for SDK and firmware compatibility (Default=True).
        :param ble_vendor_id: BLE Vendor ID for ADI dongle (Default=0x0456).
        :param ble_product_id: BLE Product ID for ADI dongle (Default=0x2CFE).
        :param ble_serial_number: BLE Serial Number ID for ADI dongle, if none then it will choose first device
         with given vendor and product ID (Default=None).
        :param ble_timeout: BLE connection timeout (Default=10).
        :param check_existing_connection: Optional argument to bypass existing connections check (Default=True).
        :param communication_mode: Selected communication modes, SerialOrDongle or BleNative
            * CommunicationMode.SerialOrDongle - serial of ble dongle will be used
            * CommunicationMode.BleNative - BleNative only. Parameters mac_address and ble_timeout  will be used.
                        Where mac_address is the MAC address of the target device
        """
        self._board = None
        self._mac_address = None
        self._ble_manager = None
        self._alarms_args = None
        self._serial_object = None
        self._alarms_callback_function = None
        self.communication_mode = communication_mode

        if self.communication_mode == CommunicationMode.BleNative and mac_address is None:
            raise ErroneousParametersCombination('BLE communication mode selected, but BLE MAC address not specified')

        self.reconnect(serial_port_address, mac_address, baud_rate, board, logging_filename, debug,
                       sync_date_time, check_version, ble_vendor_id, ble_product_id, ble_serial_number, ble_timeout,
                       check_existing_connection, **kwargs)

    def reconnect(self, serial_port_address: str, mac_address: str = None, baud_rate: int = 921600,
                  board=Board.STUDY_WATCH, logging_filename: str = None, debug: bool = False,
                  sync_date_time=False, check_version=False, ble_vendor_id: int = 0x0456, ble_product_id: int = 0x2CFE,
                  ble_serial_number: str = None, ble_timeout: int = 10, check_existing_connection: bool = True,
                  **kwargs):
        """
        reconnect method allows you to reconnect to SDK; you must call disconnect before using connect.

        :param serial_port_address: serial port of the device connected.
        :param mac_address: MAC address of the device.
        :param baud_rate: baud rate.
        :param board: board to connect (STUDY_WATCH, VSM_MB_SB).
        :param logging_filename: log file name.
        :param debug: control for debug mode.
        :param sync_date_time: Sync current system date and time with firmware (Default=False).
        :param check_version: Check for SDK and firmware compatibility (Default=False).
        :param ble_vendor_id: BLE Vendor ID for ADI dongle (Default=0x0456).
        :param ble_product_id: BLE Product ID for ADI dongle (Default=0x2CFE).
        :param ble_serial_number: BLE Serial Number ID for ADI dongle, if none then it will choose first device
         with given vendor and product ID (Default=None).
        :param ble_timeout: BLE connection timeout (Default=10).
        :param check_existing_connection: Optional argument to bypass existing connections check (Default=True).
        """

        log_format = '%(asctime)s - %(threadName)s - %(name)s - %(levelname)s - %(message)s'
        date_format = '%Y-%m-%d %H:%M:%S'
        if logging_filename:
            logging.basicConfig(filename=logging_filename, filemode='a', format=log_format, datefmt=date_format,
                                level=logging.DEBUG)
        else:
            if debug:
                logging.basicConfig(format=log_format, datefmt=date_format, level=logging.DEBUG)
            else:
                logging.basicConfig(format=log_format, datefmt=date_format)
        logger.debug("----- Study Watch SDK Started -----")
        self._mac_address = mac_address
        self._board = board
        version = None
        for _ in range(3):
            if self.communication_mode == CommunicationMode.SerialOrDongle:
                if self._mac_address:
                    candidates = list(list_ports.grep(serial_port_address))
                    if len(candidates) == 0:
                        raise DongleNotFoundError("Can't find the BLE dongle.")
                    if ble_serial_number is None:
                        ble_serial_number = candidates[0].serial_number
                    self._ble_manager = BLEManager(ble_vendor_id, ble_product_id, ble_timeout, ble_serial_number)
                    self._ble_manager.disconnect()
                    self._ble_manager.connect(self._mac_address)

                self._serial_object = SerialConnection(serial_port_address, baud_rate, **kwargs)
            else:
                self._serial_object = NrfUartBleConnection(ble_address=self._mac_address)
                self._serial_object.open(timeout=ble_timeout)

            SDK._packet_manager = PacketManager(self._serial_object)
            self._subscribe_alarms()

            if self.communication_mode == CommunicationMode.SerialOrDongle:
                if self._mac_address:
                    SDK._packet_manager.set_ble_source()
                else:
                    SDK._packet_manager.set_usb_source()
            else:
                SDK._packet_manager.set_ble_source()

            SDK._packet_manager.start_receive_and_process_threads()
            pm_app = self.get_pm_application()
            if check_existing_connection:
                tool_address = pm_app.get_existing_connection()["payload"]["tool_address"]
                if not tool_address == Application.NULL:
                    raise ConnectionAlreadyExistError(f"Connection Address: {tool_address}")
            if sync_date_time:
                pm_app.set_datetime(datetime.now())
            version = pm_app.get_version()
            if not version["payload"]["status"] == CommonStatus.NO_RESPONSE:
                break
            self.disconnect()

        if check_version:
            self._check_version(version)

    @staticmethod
    def _check_version(version):
        if version["payload"]["status"] == CommonStatus.NO_RESPONSE:
            raise Exception("Can't establish connection to the study watch.")

        supported_version = f"{version['payload']['major_version']}.{version['payload']['minor_version']}." \
                            f"{version['payload']['patch_version']}"
        if not supported_version == "5.22.0":
            logger.warning(f"Current firmware is not fully compatible with this SDK Version.\n"
                           f"Supported Firmware Version :: 5.22.0\n"
                           f"Current Firmware Version :: {supported_version}\n"
                           f"SDK Version :: {SDK.__version__}\n")

    def _subscribe_alarms(self):
        for app in [Application.ADP5360, Application.FS, Application.ECG, Application.ADPD, Application.PM, Application.LT_APP, Application.USER0_APP]:
            packet_id = self._get_packet_id(CommonCommand.ALARM_NOTIFICATION, app)
            SDK._packet_manager.subscribe(packet_id, self._alarms_callback)

    # noinspection PyUnusedLocal
    def _alarms_callback(self, data, packet_id):
        response_packet = AlarmPacket()
        response_packet.decode_packet(data)
        response_packet = response_packet.get_dict()
        if self._alarms_callback_function:
            self._alarms_callback_function(response_packet)
        else:
            logger.warning(f"ALARM - {response_packet['header']['source']} - {response_packet['payload']['status']}")

    def set_alarms_callback(self, callback_function, args=()):
        self._alarms_args = args
        self._alarms_callback_function = callback_function

    def get_adpd_application(self):
        """
        Creates an adpd application object

        :returns: an Adpd Application
        :rtype: ADPDApplication
        """
        return ADPDApplication(SDK._packet_manager)

    def get_adxl_application(self):
        """
        Creates an adxl application object

        :returns: an Adxl Application
        :rtype: ADXLApplication
        """
        return ADXLApplication(SDK._packet_manager)

    def get_adp5360_application(self):
        """
        Creates an adp5360 application object

        :returns: an ADP5360 Application
        :rtype: ADP5360Application
        """
        return ADP5360Application(SDK._packet_manager)

    def get_ecg_application(self):
        """
        Creates an ecg application object

        :returns: an ecg Application
        :rtype: ECGApplication
        """
        return ECGApplication(SDK._packet_manager)

    def get_eda_application(self):
        """
        Creates an eda application object

        :returns: an eda Application
        :rtype: EDAApplication
        """
        return EDAApplication(SDK._packet_manager)

    def get_fs_application(self):
        """
        Creates an fs application object

        :returns: an fs Application
        :rtype: FSApplication
        """
        return FSApplication(SDK._packet_manager)

    def get_pedometer_application(self):
        """
        Creates an pedometer application object

        :returns: an pedometer Application
        :rtype: PedometerApplication
        """
        return PedometerApplication(SDK._packet_manager)

    def get_pm_application(self):
        """
        Creates an pm application object

        :returns: an pm Application
        :rtype: VSMPMApplication
        """
        if self._board == Board.VSM_MB_SB:
            return VSMPMApplication(SDK._packet_manager)
        else:
            return PMApplication(SDK._packet_manager)

    def get_ppg_application(self):
        """
        Creates an ppg application object

        :returns: an Ppg Application
        :rtype: PPGApplication
        """
        return PPGApplication(SDK._packet_manager)

    def get_temperature_application(self):
        """
        Creates a temperature application object

        :returns: a Temperature Application
        :rtype: TemperatureApplication
        """
        return TemperatureApplication(SDK._packet_manager)

    def get_sqi_application(self):
        """
        Creates a sqi application object

        :returns: a SQI Application
        :rtype: SQIApplication
        """
        return SQIApplication(SDK._packet_manager)

    def get_bia_application(self):
        """
        Creates a bia application object

        :returns: a BIA Application
        :rtype: BIAApplication
        """
        return BIAApplication(SDK._packet_manager)

    def get_ad7156_application(self):
        """
        Creates a ad7156 application object

        :returns: a AD7156 Application
        :rtype: AD7156Application
        """
        return AD7156Application(SDK._packet_manager)

    def get_low_touch_application(self):
        """
        Creates a low touch application object

        :returns: a LowTouch Application
        :rtype: LowTouchApplication
        """
        return LowTouchApplication(SDK._packet_manager)

    def get_test_application(self, key_test_callback=None, cap_sense_callback=None):
        """
        Creates a test application object, used for internal firmware testing.

        :returns: a Test Application
        :rtype: TestApplication
        """
        return TestApplication(key_test_callback, cap_sense_callback, SDK._packet_manager)

    def get_user0_application(self):
        """
        Creates a User0 application object

        :returns: a User Application
        :rtype: User0Application
        """
        return User0Application(SDK._packet_manager)

    def get_lt_mode4_application(self):
        """
        Creates a lt_mode4 application object

        :returns: a lt_mode4 Application
        :rtype: LTMode4Application
        """
        return LTMode4Application(SDK._packet_manager)

    def get_session_manager_application(self):
        """
        Creates a Session Manager application object

        :returns: a Session Manager Application
        :rtype: SessionManagerApplication
        """
        return SessionManagerApplication(SDK._packet_manager)

    def unsubscribe_all_streams(self):
        """
        Unsubscribe from all application streams
        """
        result = [self.get_adxl_application().unsubscribe_stream(), self.get_sqi_application().unsubscribe_stream(),
                  self.get_ppg_application().unsubscribe_stream(), self.get_bia_application().unsubscribe_stream(),
                  self.get_ecg_application().unsubscribe_stream(), self.get_eda_application().unsubscribe_stream(),
                  self.get_temperature_application().unsubscribe_stream(),
                  self.get_pedometer_application().unsubscribe_stream()]
        adpd_app = self.get_adpd_application()
        fs_app = self.get_fs_application()
        for stream in adpd_app.get_supported_streams():
            result.append(adpd_app.unsubscribe_stream(stream))
        for stream in fs_app.get_supported_streams():
            result.append(fs_app.unsubscribe_stream(stream))
        return result

    @staticmethod
    def _get_packet_id(response_command, destination):
        return utils.join_multi_length_packets(destination.value + response_command.value)

    @staticmethod
    def get_supported_boards():
        return [SDK.STUDY_WATCH, SDK.VSM_MB_SB]

    @staticmethod
    def get_available_ports() -> List:
        """
        returns the list of tuple (port, description, hardware_id) of available ports.
        """
        result = []
        for port, desc, hardware_id in sorted(list_ports.comports()):
            result.append((port, desc, hardware_id))
        return result

    @staticmethod
    def join_csv(*args, output_filename="combined.csv", display_progress=True, progress_callback=None):
        """
        Joins multiple data stream csv file into single csv file.
        """
        file_size = os.path.getsize(args[0])
        current_size = 0
        progress_bar = None
        if display_progress:
            progress_bar = tqdm(total=file_size)
            progress_bar.set_description("ADPD csv join")
        space = {}
        header = []
        all_csv = {}
        first_iter = {}
        result_file = open(output_filename, 'w')

        for file in args:
            csv_file = open(file, 'r')
            header = [csv_file.readline(), csv_file.readline()]
            all_csv[file] = csv_file
            first_iter[file] = True
            space[file] = 0

        # write header
        result_file.write(header[0])
        result_file.write(header[1])

        while True:
            result_line = ""
            empty_file_count = 0
            for file in args:
                line = all_csv[file].readline().strip()
                # checking for empty file condition
                if len(line) == 0:
                    empty_file_count += 1
                    result_line += "," * space[file]
                    continue
                # saving number of elements in a row
                if first_iter[file]:
                    space[file] = len(line.split(","))
                    first_iter[file] = False
                # concat row
                result_line += line + ","
                # progress bar
                if file == args[0]:
                    line_size = len(line.encode('utf-8'))
                    current_size += line_size
                    if display_progress:
                        progress_bar.update(line_size)
                    if progress_callback:
                        progress_callback(SDK.JOINING_CSV, file_size, current_size)
            # break condition
            if empty_file_count == len(args):
                break
            result_file.write(result_line + "\n")

        # cleanup
        result_file.close()
        if display_progress:
            progress_bar.update(file_size - current_size)
            progress_bar.close()
        if progress_callback:
            progress_callback(SDK.JOINING_CSV, file_size, file_size)
        for file in args:
            all_csv[file].close()

    @staticmethod
    def _csv_write_config(file, config, header):
        file.write(f"Address, Value, {header}\n")
        if not type(config) == list:
            config = [config]
        for data in config:
            for value in data["payload"]["data"]:
                if header == "#PPG_LCFG":
                    file.write(f"{value}\n")
                else:
                    file.write(f"{value[0]}, {value[1]}\n")
        file.write(f"\n")

    @staticmethod
    def _csv_write_version(file, config, header):
        file.write(f"Module, {header}\n")
        file.write(f"Major version, {config['payload']['major_version']}\n")
        file.write(f"Minor version, {config['payload']['minor_version']}\n")
        file.write(f"Patch version, {config['payload']['patch_version']}\n")
        file.write(f"Version string, {config['payload']['version_string']}\n")
        file.write(f"Build version, {config['payload']['build_version']}\n")
        file.write(f"\n")

    @staticmethod
    def _packet_loss(file, msg, packet_count):
        if packet_count:
            file.write(f"{msg}, {packet_count}\n")
        else:
            file.write(f"{msg}, 0\n")

    # noinspection PyProtectedMember,PyTypeChecker,PyUnresolvedReferences
    @staticmethod
    def convert_log_to_csv(filename, display_progress=True, progress_callback=None):
        """
        Converts M2M2 log file into csv.
        """
        if not os.path.exists(filename):
            raise Exception("File not Found.")

        info_result = {
            "key_value_pair": None, "datetime": None, "system_info": None, "version": None, "ppg_algo_version": None,
            "ped_algo_version": None, "ecg_algo_version": None, "sqi_algo_version": None, "adpd_dcfg": None,
            "adxl_dcfg": None, "ppg_lcfg": None, "ecg_lcfg": None,
        }
        folder_name = os.path.splitext(filename)[0]
        try:
            os.mkdir(folder_name)
        except Exception as e:
            logger.debug(e)

        # creating all applications
        packet_manager = PacketManager(None, filename=filename)
        packet_manager.set_usb_source()
        ad7156_app = AD7156Application(packet_manager)
        adpd_app = ADPDApplication(packet_manager)
        adxl_app = ADXLApplication(packet_manager)
        ecg_app = ECGApplication(packet_manager)
        eda_app = EDAApplication(packet_manager)
        ped_app = PedometerApplication(packet_manager)
        ppg_app = PPGApplication(packet_manager)
        temp_app = TemperatureApplication(packet_manager)
        sqi_app = SQIApplication(packet_manager)
        bia_app = BIAApplication(packet_manager)
        pm_app = PMApplication(packet_manager)
        adp5360_app = ADP5360Application(packet_manager)

        # enabling csv logging
        adxl_app.enable_csv_logging(f"{folder_name}/adxl.csv")
        for i, stream in enumerate(adpd_app.get_supported_streams()):
            if stream == adpd_app.STREAM_STATIC_AGC:
                adpd_app.enable_csv_logging(f"{folder_name}/static_agc.csv", stream=adpd_app.STREAM_STATIC_AGC)
            else:
                adpd_app.enable_csv_logging(f"{folder_name}/adpd{i + 1}.csv", stream=stream)
        ecg_app.enable_csv_logging(f"{folder_name}/ecg.csv")
        eda_app.enable_csv_logging(f"{folder_name}/eda.csv")
        ped_app.enable_csv_logging(f"{folder_name}/ped.csv")
        ppg_app.enable_csv_logging(f"{folder_name}/ppg.csv", stream=ppg_app.STREAM_PPG)
        ppg_app.enable_csv_logging(f"{folder_name}/sync_ppg.csv", stream=ppg_app.STREAM_SYNC_PPG)
        ppg_app.enable_csv_logging(f"{folder_name}/dynamic_agc.csv", stream=ppg_app.STREAM_DYNAMIC_AGC)
        ppg_app.enable_csv_logging(f"{folder_name}/hrv.csv", stream=ppg_app.STREAM_HRV)
        for i, stream in enumerate(temp_app.get_supported_streams()):
            temp_app.enable_csv_logging(f"{folder_name}/temperature{i + 1}.csv", stream=stream)
        sqi_app.enable_csv_logging(f"{folder_name}/sqi.csv")
        bia_app.enable_csv_logging(f"{folder_name}/bia.csv", stream=bia_app.STREAM_BIA)
        bia_app.enable_csv_logging(f"{folder_name}/bcm.csv", stream=bia_app.STREAM_BCM)
        adp5360_app.enable_csv_logging(f"{folder_name}/adp.csv")
        ad7156_app.enable_csv_logging(f"{folder_name}/ad7156.csv")

        # subscribing
        apps = [adxl_app, ecg_app, eda_app, ped_app, sqi_app, ad7156_app]
        for app in apps:
            app._subscribe_stream_data()
        ppg_app._subscribe_stream_data(ppg_app.STREAM_PPG)
        ppg_app._subscribe_stream_data(ppg_app.STREAM_SYNC_PPG)
        ppg_app._subscribe_stream_data(ppg_app.STREAM_DYNAMIC_AGC)
        ppg_app._subscribe_stream_data(ppg_app.STREAM_HRV)
        adp5360_app._subscribe_stream_data(adp5360_app.STREAM_BATTERY)
        bia_app._subscribe_stream_data(bia_app.STREAM_BIA)
        bia_app._subscribe_stream_data(bia_app.STREAM_BCM)
        for stream in adpd_app.get_supported_streams():
            adpd_app._subscribe_stream_data(stream)
        for stream in temp_app.get_supported_streams():
            temp_app._subscribe_stream_data(stream)

        def update_info_dict(packet, response_byte, key):
            packet.decode_packet(response_byte)
            if info_result.get(key) is None:
                info_result[key] = packet.get_dict()
            else:
                if type(info_result[key]) == list:
                    info_result[key] = info_result[key] + [packet.get_dict()]
                else:
                    info_result[key] = [info_result[key], packet.get_dict()]

        # noinspection PyShadowingNames,PyProtectedMember
        def callback(response_byte, packet_id):
            if SDK._get_packet_id(FSCommand.SET_KEY_VALUE_PAIR_RES, Application.FS) == packet_id:
                packet = KeyValuePairResponsePacket()
                update_info_dict(packet, response_byte, "key_value_pair")
            elif SDK._get_packet_id(PMCommand.GET_DATE_TIME_RES, Application.PM) == packet_id:
                packet = DateTimePacket()
                update_info_dict(packet, response_byte, "datetime")
                response_packet = info_result["datetime"]
                tz_sec = response_packet["payload"]["tz_sec"]
                offset = timezone(timedelta(seconds=tz_sec))
                dt = datetime(response_packet['payload']['year'], response_packet['payload']['month'],
                              response_packet['payload']['day'], response_packet['payload']['hour'],
                              response_packet['payload']['minute'], response_packet['payload']['second'],
                              tzinfo=offset)
                for _stream in adpd_app.get_supported_streams():
                    adpd_app._update_timestamp(dt, _stream, generate_ts=True, tz_sec=tz_sec)
                for _stream in temp_app.get_supported_streams():
                    temp_app._update_timestamp(dt, _stream, generate_ts=True, tz_sec=tz_sec)
                for _app in apps:
                    _app._update_timestamp(dt, generate_ts=True, tz_sec=tz_sec)
                ppg_app._update_timestamp(dt, ppg_app.STREAM_PPG, generate_ts=True, tz_sec=tz_sec)
                ppg_app._update_timestamp(dt, ppg_app.STREAM_SYNC_PPG, generate_ts=True, tz_sec=tz_sec)
                ppg_app._update_timestamp(dt, ppg_app.STREAM_DYNAMIC_AGC, generate_ts=True, tz_sec=tz_sec)
                ppg_app._update_timestamp(dt, ppg_app.STREAM_HRV, generate_ts=True, tz_sec=tz_sec)
                bia_app._update_timestamp(dt, bia_app.STREAM_BIA, generate_ts=True, tz_sec=tz_sec)
                bia_app._update_timestamp(dt, bia_app.STREAM_BCM, generate_ts=True, tz_sec=tz_sec)
                adp5360_app._update_timestamp(dt, adp5360_app.STREAM_BATTERY, generate_ts=True, tz_sec=tz_sec)

            elif SDK._get_packet_id(PMCommand.SYS_INFO_RES, Application.PM) == packet_id:
                packet = SystemInfoPacket()
                update_info_dict(packet, response_byte, "system_info")
            elif SDK._get_packet_id(CommonCommand.GET_VERSION_RES, Application.PM) == packet_id:
                packet = VersionPacket()
                update_info_dict(packet, response_byte, "version")
            elif SDK._get_packet_id(PPGCommand.GET_ALGO_VENDOR_VERSION_RES, Application.PPG) == packet_id:
                packet = VersionPacket()
                update_info_dict(packet, response_byte, "ppg_algo_version")
            elif SDK._get_packet_id(PedometerCommand.GET_ALGO_VENDOR_VERSION_RES, Application.PEDOMETER) == packet_id:
                packet = VersionPacket()
                update_info_dict(packet, response_byte, "ped_algo_version")
            elif SDK._get_packet_id(ECGCommand.GET_ALGO_VENDOR_VERSION_RES, Application.ECG) == packet_id:
                packet = VersionPacket()
                update_info_dict(packet, response_byte, "ecg_algo_version")
            elif SDK._get_packet_id(SQICommand.GET_ALGO_VENDOR_VERSION_RES, Application.SQI) == packet_id:
                packet = VersionPacket()
                update_info_dict(packet, response_byte, "sqi_algo_version")
            elif SDK._get_packet_id(CommonCommand.GET_DCFG_RES, Application.ADXL) == packet_id:
                packet = ADXLDCFGPacket()
                update_info_dict(packet, response_byte, "adxl_dcfg")
            elif SDK._get_packet_id(CommonCommand.GET_DCFG_RES, Application.ADPD) == packet_id:
                packet = ADPDDCFGPacket()
                update_info_dict(packet, response_byte, "adpd_dcfg")
            elif SDK._get_packet_id(CommonCommand.GET_LCFG_RES, Application.PPG) == packet_id:
                packet = LibraryConfigDataPacket()
                update_info_dict(packet, response_byte, "ppg_lcfg")
            elif SDK._get_packet_id(CommonCommand.READ_LCFG_RES, Application.ECG) == packet_id:
                packet = ECGLibraryConfigPacket()
                update_info_dict(packet, response_byte, "ecg_lcfg")

        initial_packets = [
            [FSCommand.SET_KEY_VALUE_PAIR_RES, Application.FS],
            [PMCommand.GET_DATE_TIME_RES, Application.PM],
            [PMCommand.SYS_INFO_RES, Application.PM],
            [CommonCommand.GET_VERSION_RES, Application.PM],
            [PPGCommand.GET_ALGO_VENDOR_VERSION_RES, Application.PPG],
            [PedometerCommand.GET_ALGO_VENDOR_VERSION_RES, Application.PEDOMETER],
            [ECGCommand.GET_ALGO_VENDOR_VERSION_RES, Application.ECG],
            [SQICommand.GET_ALGO_VENDOR_VERSION_RES, Application.SQI],
            [CommonCommand.GET_DCFG_RES, Application.ADXL],
            [CommonCommand.GET_DCFG_RES, Application.ADPD],
            [CommonCommand.GET_LCFG_RES, Application.PPG],
            [CommonCommand.READ_LCFG_RES, Application.ECG]
        ]
        for initial_packet in initial_packets:
            packet_id = SDK._get_packet_id(initial_packet[0], initial_packet[1])
            packet_manager.subscribe(packet_id, callback)

        # process file
        packet_manager.process_file(display_progress, progress_callback)

        # disabling csv logging.
        for stream in adpd_app.get_supported_streams():
            adpd_app.disable_csv_logging(stream=stream)
        for stream in temp_app.get_supported_streams():
            temp_app.disable_csv_logging(stream=stream)
        for app in apps:
            app.disable_csv_logging()
        ppg_app.disable_csv_logging(stream=ppg_app.STREAM_PPG)
        ppg_app.disable_csv_logging(stream=ppg_app.STREAM_SYNC_PPG)
        ppg_app.disable_csv_logging(stream=ppg_app.STREAM_DYNAMIC_AGC)
        ppg_app.disable_csv_logging(stream=ppg_app.STREAM_HRV)
        bia_app.disable_csv_logging(stream=bia_app.STREAM_BIA)
        bia_app.disable_csv_logging(stream=bia_app.STREAM_BCM)
        adp5360_app.disable_csv_logging()

        # unsubscribing
        for app in apps:
            app._unsubscribe_stream_data()
        ppg_app._unsubscribe_stream_data(ppg_app.STREAM_PPG)
        ppg_app._unsubscribe_stream_data(ppg_app.STREAM_SYNC_PPG)
        ppg_app._unsubscribe_stream_data(ppg_app.STREAM_DYNAMIC_AGC)
        ppg_app._unsubscribe_stream_data(ppg_app.STREAM_HRV)
        bia_app._unsubscribe_stream_data(bia_app.STREAM_BIA)
        bia_app._unsubscribe_stream_data(bia_app.STREAM_BCM)
        adp5360_app._unsubscribe_stream_data(adp5360_app.STREAM_BATTERY)
        for stream in adpd_app.get_supported_streams():
            adpd_app._unsubscribe_stream_data(stream)
        for stream in temp_app.get_supported_streams():
            temp_app._unsubscribe_stream_data(stream)

        with open(f"{folder_name}/Summary.csv", 'w') as f:
            if info_result['key_value_pair']:
                f.write(f"Participant Information, {info_result['key_value_pair']['payload']['value_id']}" + "\n")
                # packet loss check
                SDK._packet_loss(f, "ADXL Packet loss", adxl_app.get_packet_lost_count())
                for i, stream in enumerate(adpd_app.get_supported_streams()):
                    if stream == adpd_app.STREAM_STATIC_AGC:
                        SDK._packet_loss(f, "STATIC AGC Packet loss", adpd_app.get_packet_lost_count(stream))
                    else:
                        adpd_lost_packets = adpd_app.get_packet_lost_count(stream)
                        if adpd_lost_packets:
                            SDK._packet_loss(f, f"ADPD{i + 1} CH1 Packet loss", adpd_lost_packets[0])
                            SDK._packet_loss(f, f"ADPD{i + 1} CH2 Packet loss", adpd_lost_packets[1])
                        else:
                            SDK._packet_loss(f, f"ADPD{i + 1} CH1 Packet loss", 0)
                            SDK._packet_loss(f, f"ADPD{i + 1} CH2 Packet loss", 0)
                for i, stream in enumerate(temp_app.get_supported_streams()):
                    SDK._packet_loss(f, f"Temperature{i + 1} Packet loss", temp_app.get_packet_lost_count(stream))

                SDK._packet_loss(f, "AD7156 Packet loss", ad7156_app.get_packet_lost_count())
                SDK._packet_loss(f, "ECG Packet loss", ecg_app.get_packet_lost_count())
                SDK._packet_loss(f, "EDA Packet loss", eda_app.get_packet_lost_count())
                SDK._packet_loss(f, "Pedometer Packet loss", ped_app.get_packet_lost_count())
                SDK._packet_loss(f, "PPG Packet loss", ppg_app.get_packet_lost_count(ppg_app.STREAM_PPG))
                SDK._packet_loss(f, "SYNC PPG Packet loss", ppg_app.get_packet_lost_count(ppg_app.STREAM_SYNC_PPG))
                SDK._packet_loss(f, "DYNAMIC AGC Packet loss",
                                 ppg_app.get_packet_lost_count(ppg_app.STREAM_DYNAMIC_AGC))
                SDK._packet_loss(f, "HRV Packet loss", ppg_app.get_packet_lost_count(ppg_app.STREAM_HRV))
                SDK._packet_loss(f, "SQI Packet loss", sqi_app.get_packet_lost_count())
                SDK._packet_loss(f, "BIA Packet loss", bia_app.get_packet_lost_count(bia_app.STREAM_BIA))
                SDK._packet_loss(f, "BCM Packet loss", bia_app.get_packet_lost_count(bia_app.STREAM_BCM))

        with open(f"{folder_name}/Configuration.csv", 'w') as f:
            if info_result["adpd_dcfg"]:
                SDK._csv_write_config(f, info_result["adpd_dcfg"], "#ADPD_DCFG")
            if info_result["adxl_dcfg"]:
                SDK._csv_write_config(f, info_result["adxl_dcfg"], "#ADXL_DCFG")
            if info_result["ppg_lcfg"]:
                SDK._csv_write_config(f, info_result["ppg_lcfg"], "#PPG_LCFG")
            if info_result["ecg_lcfg"]:
                SDK._csv_write_config(f, info_result["ecg_lcfg"], "#ECG_LCFG")

        with open(f"{folder_name}/FirmwareVersion.csv", 'w') as f:
            if info_result["version"]:
                SDK._csv_write_version(f, info_result["version"], "PM Firmware Version")
            if info_result["ppg_algo_version"]:
                SDK._csv_write_version(f, info_result["ppg_algo_version"], "PPG Firmware Version")
            if info_result["ped_algo_version"]:
                SDK._csv_write_version(f, info_result["ped_algo_version"], "PED Firmware Version")
            if info_result["ecg_algo_version"]:
                SDK._csv_write_version(f, info_result["ecg_algo_version"], "ECG Firmware Version")
            if info_result["sqi_algo_version"]:
                SDK._csv_write_version(f, info_result["sqi_algo_version"], "SQI Firmware Version")

        with open(f"{folder_name}/PSBoardInfo.csv", 'w') as f:
            if info_result["system_info"]:
                config = info_result["system_info"]
                f.write(f"Module, PS Board Info\n")
                f.write(f"Version, {config['payload']['version']}\n")
                f.write(f"MAC address, {config['payload']['mac_address']}\n")
                f.write(f"Device ID, {config['payload']['device_id']}\n")
                f.write(f"Model number, {config['payload']['model_number']}\n")
                f.write(f"Hardware ID, {config['payload']['hw_id']}\n")
                f.write(f"BOM ID, {config['payload']['bom_id']}\n")
                f.write(f"Batch ID, {config['payload']['batch_id']}\n")
                f.write(f"Board Type, {config['payload']['board_type']}\n")

        # combining adpd csv
        adpd_csv_files = []
        for i in range(1, 13):
            adpd_file = f"{folder_name}/adpd{i}.csv"
            if os.path.exists(adpd_file):
                adpd_csv_files.append(adpd_file)
        if len(adpd_csv_files):
            SDK.join_csv(*adpd_csv_files, output_filename=f"{folder_name}/adpd_streams.csv",
                         display_progress=display_progress, progress_callback=progress_callback)

        for initial_packet in initial_packets:
            packet_id = SDK._get_packet_id(initial_packet[0], initial_packet[1])
            packet_manager.unsubscribe(packet_id, callback)

        # resetting packet manager
        AD7156Application(SDK._packet_manager)
        ADPDApplication(SDK._packet_manager)
        ADXLApplication(SDK._packet_manager)
        ECGApplication(SDK._packet_manager)
        EDAApplication(SDK._packet_manager)
        PedometerApplication(SDK._packet_manager)
        PPGApplication(SDK._packet_manager)
        TemperatureApplication(SDK._packet_manager)
        SQIApplication(SDK._packet_manager)
        BIAApplication(SDK._packet_manager)
        PMApplication(SDK._packet_manager)
        ADP5360Application(SDK._packet_manager)

    def convert_ticks_to_timestamp(self, ticks: int):
        """
        Converts firmware ticks to unix timestamp.
        """
        date_time_packet = self.get_pm_application().get_datetime()
        offset = timezone(timedelta(seconds=date_time_packet["payload"]["tz_sec"]))
        date_time = datetime(date_time_packet['payload']['year'], date_time_packet['payload']['month'],
                             date_time_packet['payload']['day'], date_time_packet['payload']['hour'],
                             date_time_packet['payload']['minute'], date_time_packet['payload']['second'],
                             tzinfo=offset)
        ts = (32000.0 * ((date_time.hour * 3600) + (date_time.minute * 60) + date_time.second))
        ref_time = date_time.timestamp()
        change = ts - ticks
        change = change / 32000.0
        return (ref_time - change) * 1000

    def disconnect(self):
        """disconnect SDK"""
        logger.debug("----- Study Watch SDK Stopped -----")

        if self.communication_mode == CommunicationMode.SerialOrDongle and self._mac_address is not None:
            self._ble_manager.disconnect()

        SDK._packet_manager.close()
        time.sleep(1)

    def is_connected(self):

        if self.communication_mode == CommunicationMode.SerialOrDongle:
            if self._ble_manager is not None:
                return self._ble_manager.is_connected()

        return self._serial_object.is_connected()



import logging
import threading
import time
from queue import Queue
from sys import platform

import usb1
from . import utils

logger = logging.getLogger(__name__)


class BLEManager:
    RID_CMD = 0x01
    RID_RSP = 0x02
    RID_NOT = 0x03

    CMD_SCAN_STOP = 0x01
    CMD_SCAN_START = 0x00
    CMD_CONNECT = 0x02
    CMD_DISCONNECT = 0x03
    CMD_RESET = 0x7F

    MAX_LENGTH = 63

    def __init__(self, vendor_id, product_id, timeout=5, dongle_serial_number=None):
        self.queue = Queue()
        self.timeout = timeout
        self.vendor_id = vendor_id
        self.product_id = product_id
        self.dongle_serial_number = dongle_serial_number
        self.device = None
        self._is_connected: threading.Event = threading.Event()

    def _open(self):
        context = usb1.USBContext()
        device_list = context.getDeviceList(skip_on_error=True)
        device = None
        for dev in device_list:
            try:
                pid = dev.getProductID()
                vid = dev.getVendorID()
                s_number = dev.getSerialNumber()
                if vid == self.vendor_id and pid == self.product_id:
                    if s_number == self.dongle_serial_number:
                        device = dev
                        break
            except:
                pass
        if device is None:
            raise Exception(f"Can't find BLE dongle with vendor_id={self.vendor_id}, "
                            f"product_id={self.product_id} and serial_number={self.dongle_serial_number}.")
        else:
            self.device = device
            self.device = self.device.open()
            self.device.resetDevice()
            if platform == "linux" or platform == "linux2":
                if self.device.kernelDriverActive(0):
                    self.device.detachKernelDriver(0)
            elif platform == "darwin":
                pass
            elif platform == "win32":
                pass
            self.device.claimInterface(0)
        threading.Thread(target=self.receive_thread, daemon=True).start()

    def receive_thread(self):
        while True:
            try:
                data = self.device.bulkRead(0x81, 64)
                logger.debug(f"BLE RX :: {':'.join(utils.convert_int_array_to_hex(data))}")
                self.queue.put(data)
            except Exception as e:
                break

    def _send(self, packet):
        packet = packet + [0 for _ in range(self.MAX_LENGTH - len(packet))]
        logger.debug(f"BLE TX :: {':'.join(utils.convert_int_array_to_hex(packet))}")
        self.device.bulkWrite(1, packet)

    def _scan_start(self, mac_address):
        default_name = "Nordic_USBD_BLE_UART"
        msg = [self.RID_CMD, self.CMD_SCAN_START, 0, 1] + [ord(i) for i in default_name]
        self._send(msg)
        try:
            while True:
                data = self.queue.get(timeout=self.timeout)
                logger.debug(f"Scan Result :: {'-'.join(['%02X' % i for i in data[4:10]])}")
                if data[0] == 3 and list(data[4:10]) == mac_address:
                    break
        except Exception as e:
            logger.debug(e)
            raise Exception(f"Failed to find BLE device {'-'.join(['%X' % i for i in mac_address])}.")

    def _scan_stop(self):
        msg = [self.RID_CMD, self.CMD_SCAN_STOP, 0]
        self._send(msg)
        try:
            while True:
                data = self.queue.get(timeout=self.timeout)
                if data[0] == 2:
                    break
        except Exception as e:
            logger.debug(e)
            raise Exception(f"Failed to stop scan.")

    def connect(self, mac_address):
        self._open()
        mac_address = list(map(int, mac_address.split("-"), [16 for _ in mac_address]))
        mac_address = list(reversed(mac_address))
        self._reset()
        self._open()
        self._scan_start(mac_address)
        self._scan_stop()
        msg = [self.RID_CMD, self.CMD_CONNECT] + mac_address
        self._send(msg)
        try:
            while True:
                data = self.queue.get(timeout=self.timeout)
                if data[0] == 3:
                    time.sleep(2)
                    break
        except Exception as e:
            logger.debug(e)
            raise Exception(f"Can't connect to BLE {'-'.join(['%X' % i for i in mac_address])}.")
        logger.info(f"BLE connected to {'-'.join(['%X' % i for i in mac_address])}.")
        self._is_connected.set()

    def disconnect(self):
        self._is_connected.clear()
        self._open()
        msg = [self.RID_CMD, self.CMD_DISCONNECT]
        self._send(msg)
        time.sleep(1)
        self.queue.empty()
        self.device.close()

    def _reset(self):
        msg = [self.RID_CMD, self.CMD_RESET, 0]
        self._send(msg)
        time.sleep(5)

    def is_connected(self) -> bool:
        return self._is_connected.is_set()




import csv
import math
import logging
import time
from datetime import datetime, timezone, timedelta

from ..core.enums.common_enums import Stream

logger = logging.getLogger(__name__)


class CSVLogger:
    def __init__(self, filename, header, write_header=True):
        self.ch1 = []
        self.ch2 = []
        self.file = None
        self.ch1_dark = []
        self.ch2_dark = []
        self.ch1_time = []
        self.ch2_time = []
        self.writer = None
        self.header = header
        self.filename = filename
        self.logging_started = None
        self.write_header = write_header

    def write_row(self, row):
        if self.writer:
            self.writer.writerow(row)

    def start_logging(self, last_timestamp, tz_sec):
        try:
            self.file = open(self.filename, 'w', newline="")
            self.writer = csv.writer(self.file, quoting=csv.QUOTE_NONE)
            offset = timezone(timedelta(seconds=tz_sec))
            dt_object = datetime.fromtimestamp(last_timestamp, tz=offset)
            self.write_row(["Local: ", dt_object, "tz_sec ", tz_sec])
            self.write_row(["Linux Epoch (ms):", last_timestamp * 1000])
            if self.write_header:
                self.write_row(self.header)
        except Exception as e:
            logger.error(f"Error while opening the {self.filename} file, reason :: {e}.", exc_info=True)

    def stop_logging(self):
        data_len = max(len(self.ch1_time), len(self.ch2_time))
        if len(self.ch1_time) > 0 and len(self.ch2_time) > 0:
            if not len(self.ch1_time) == len(self.ch2_time):
                data_len = min(len(self.ch1_time), len(self.ch2_time))
                logger.warning(f"Issue with ADPD CSV logging. Recommended sequence unsubscribe_stream "
                               f"then disable_csv_logging.")
        for i in range(data_len):
            if len(self.ch2) == 0:
                self.adpd_add_row([self.ch1_time[i], self.ch1_dark[i], self.ch1[i]], 1)
            elif len(self.ch1) == 0:
                self.adpd_add_row([self.ch2_time[i], self.ch2_dark[i], self.ch2[i]], 2)
            else:
                self.adpd_add_row(
                    [self.ch1_time[i], self.ch1_dark[i], self.ch1[i], self.ch2_dark[i], self.ch2[i]], 0)

        if self.file:
            self.file.close()

    def add_row(self, result, last_timestamp, tz_sec):
        stream = result["header"]["source"]
        if self.logging_started is None:
            self.logging_started = True
            self.start_logging(last_timestamp, tz_sec)
        if stream == Stream.ADXL:
            self._adxl_callback(result)
        elif stream in [Stream.ADPD1, Stream.ADPD2, Stream.ADPD3, Stream.ADPD4, Stream.ADPD5, Stream.ADPD6,
                        Stream.ADPD7, Stream.ADPD8, Stream.ADPD9, Stream.ADPD10, Stream.ADPD11, Stream.ADPD12]:
            self._adpd_callback(result)
        elif stream == Stream.PPG:
            self._ppg_callback(result)
        elif stream == Stream.SYNC_PPG:
            self._syncppg_callback(result)
        elif stream == Stream.ECG:
            self._ecg_callback(result)
        elif stream == Stream.EDA:
            self._eda_callback(result)
        elif stream in [Stream.TEMPERATURE1, Stream.TEMPERATURE2, Stream.TEMPERATURE3,
                        Stream.TEMPERATURE4, Stream.TEMPERATURE5, Stream.TEMPERATURE6, Stream.TEMPERATURE7,
                        Stream.TEMPERATURE8, Stream.TEMPERATURE9, Stream.TEMPERATURE10,
                        Stream.TEMPERATURE11, Stream.TEMPERATURE12]:
            self._temp_callback(result)
        elif stream == Stream.PEDOMETER:
            self._ped_callback(result)
        elif stream == Stream.BIA:
            self._bia_callback(result)
        elif stream == Stream.BCM:
            self._bcm_callback(result)
        elif stream == Stream.SQI:
            self._sqi_callback(result)
        elif stream == Stream.BATTERY:
            self._battery_callback(result)
        elif stream in [Stream.STATIC_AGC_STREAM, Stream.DYNAMIC_AGC_STREAM]:
            self._agc_callback(result)
        elif stream == Stream.HRV:
            self._hrv_callback(result)
        elif stream == Stream.AD7156:
            self._ad7156_callback(result)

    def adpd_add_row(self, row, header_format):
        if not self.write_header:
            self.write_header = True
            self.write_row([" ", self.header[0], "", "", "", ""])
            row2 = []
            if header_format == 0:
                row2 = [" ", "CH1", "", "", "CH2", ""]
            elif header_format == 1:
                row2 = [" ", "CH1", "", ""]
            elif header_format == 2:
                row2 = [" ", "CH2", "", ""]
            self.write_row(row2)
            row3 = []
            if header_format == 0:
                row3 = [" ", "Timestamp", "D1", "S1", "D2", "S2"]
            elif header_format == 1:
                row3 = [" ", "Timestamp", "D1", "S1"]
            elif header_format == 2:
                row3 = [" ", "Timestamp", "D2", "S2"]
            self.write_row(row3)
        self.write_row([" "] + row)

    def _adxl_callback(self, data):
        for value in data["payload"]["stream_data"]:
            self.write_row([value["timestamp"], value["x"], value["y"], value["z"]])

    def _adpd_callback(self, data):
        if len(self.ch1_time) > 200 or len(self.ch2_time) > 200:
            adpd_data_length = len(data["payload"]["signal_data"])
            for i in range(adpd_data_length):
                if len(self.ch2) == 0:
                    self.adpd_add_row([self.ch1_time[i], self.ch1_dark[i], self.ch1[i]], 1)
                elif len(self.ch1) == 0:
                    self.adpd_add_row([self.ch2_time[i], self.ch2_dark[i], self.ch2[i]], 2)
                else:
                    self.adpd_add_row(
                        [self.ch1_time[i], self.ch1_dark[i], self.ch1[i], self.ch2_dark[i], self.ch2[i]], 0)

            del self.ch1[:adpd_data_length]
            del self.ch2[:adpd_data_length]
            del self.ch1_dark[:adpd_data_length]
            del self.ch2_dark[:adpd_data_length]
            del self.ch1_time[:adpd_data_length]
            del self.ch2_time[:adpd_data_length]

        if data["payload"]["channel_num"] == 1:
            for signal_data in data["payload"]["signal_data"]:
                self.ch1_time.append(data["payload"]["timestamp"])
                self.ch1.append(signal_data)
            for dark_data in data["payload"]["dark_data"]:
                self.ch1_dark.append(dark_data)

        elif data["payload"]["channel_num"] == 2:
            for adpd_data in data["payload"]["signal_data"]:
                self.ch2_time.append(data["payload"]["timestamp"])
                self.ch2.append(adpd_data)
            for dark_data in data["payload"]["dark_data"]:
                self.ch2_dark.append(dark_data)

    def _ecg_callback(self, data):
        for value in data["payload"]["stream_data"]:
            self.write_row([value["timestamp"], data["payload"]["sequence_number"], value["ecg_data"]])

    def _ppg_callback(self, data):
        hr = float(data["payload"]["hr"]) / float(16)
        confidence = (float(100) * (data["payload"]["confidence"] / float(1024)))
        self.write_row([data["payload"]["timestamp"], hr, confidence, data["payload"]["hr_type"]])

    def _syncppg_callback(self, data):
        for value in data["payload"]["stream_data"]:
            self.write_row(
                [value["ppg_timestamp"], value["ppg_data"], value["adxl_timestamp"], value["adxl_x"],
                 value["adxl_y"], value["adxl_z"]])

    def _eda_callback(self, data):
        for value in data["payload"]["stream_data"]:
            eda_real = value["real"]
            eda_imaginary = value["imaginary"]
            if eda_real == 0:
                eda_real = 1
            impedance_img = eda_imaginary * 1000
            impedance_real = eda_real * 1000
            real_and_img = float(impedance_real * impedance_real + impedance_img * impedance_img)
            impedance_module = math.sqrt(real_and_img)
            impedance_phase = math.atan2(impedance_img, impedance_real)
            admittance_real = float(impedance_real /
                                    float(impedance_real * impedance_real + impedance_img * impedance_img))
            admittance_img = -float(impedance_img /
                                    float(impedance_real * impedance_real + impedance_img * impedance_img))
            admittance_module = 1 / impedance_module
            admittance_phase = math.atan2(admittance_img, admittance_real)

            self.write_row([
                value["timestamp"], impedance_real, impedance_img, impedance_module, impedance_phase, admittance_real,
                admittance_img, admittance_module, admittance_phase, data["payload"]["sequence_number"]
            ])

    def _temp_callback(self, data):
        self.write_row([data["payload"]["timestamp"], data["payload"]["skin_temperature"],
                        data["payload"]["impedance"], data["payload"]["compensated_temperature"]])

    def _ped_callback(self, data):
        self.write_row([data["payload"]["timestamp"], data["payload"]["steps"]])

    def _bia_callback(self, data):
        for value in data["payload"]["stream_data"]:
            bcm_real = value["real"]
            bcm_imaginary = value["imaginary"]
            if bcm_real == 0:
                bcm_real = 1
            if bcm_imaginary == 0:
                bcm_imaginary = 1
            impedance_img = bcm_imaginary / 1000
            impedance_real = bcm_real / 1000
            real_and_img = float(impedance_real * impedance_real + impedance_img * impedance_img)
            impedance_module = math.sqrt(real_and_img)
            impedance_phase = math.atan2(impedance_img, impedance_real)
            admittance_real = float(impedance_real /
                                    float(impedance_real * impedance_real + impedance_img * impedance_img))
            admittance_img = -float(impedance_img /
                                    float(impedance_real * impedance_real + impedance_img * impedance_img))
            admittance_module = 1 / impedance_module
            admittance_phase = math.atan2(admittance_img, admittance_real)
            self.write_row([
                value["timestamp"], impedance_real, impedance_img, impedance_module, impedance_phase,
                admittance_real, admittance_img, admittance_module, admittance_phase,
                data["payload"]["sequence_number"], value["frequency_index"]
            ])

    def _bcm_callback(self, data):
        self.write_row([data["payload"]["timestamp"], data["payload"]["ffm_estimated"],
                        data["payload"]["bmi"], data["payload"]["fat_percent"], data["payload"]["sequence_num"]])

    def _sqi_callback(self, data):
        self.write_row([data["payload"]["timestamp"], data["payload"]["sqi"]])

    def _battery_callback(self, data):
        self.write_row([data["payload"]["timestamp"], data["payload"]["battery_status"],
                        data["payload"]["adp5360_battery_level"], data["payload"]["custom_battery_level"], data["payload"]["battery_mv"]])

    def _agc_callback(self, data):
        self.write_row([data["payload"]["timestamp"], *data["payload"]["mts"], *data["payload"]["setting"]])

    def _hrv_callback(self, data):
        for value in data["payload"]["stream_data"]:
            self.write_row([value["timestamp"], value["rr_interval"], value["is_gap"], value["rmssd"]])

    def _ad7156_callback(self, data):
        for value in data["payload"]["stream_data"]:
            self.write_row([value["timestamp"], value["ch1_cap"], value["ch2_cap"], value["ch1_ADCCode"],
                            value["ch2_ADCCode"], value["OUT1_val"], value["OUT2_val"]])



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

from unittest import TestCase

from . import common_test
from sdk_pure_python.adi_study_watch import SDK
from sdk_pure_python.adi_study_watch.core.enums.common_enums import CommonStatus


def callback(data):
    print(data)


class TestEDAApplication(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.application = SDK("COM4", debug=True).get_eda_application()
        cls.application.set_callback(callback)
        cls.application.set_timeout(30)

    # def test_calibrate_resistor_tia(self):
    #     x = self.application.calibrate_resistor_tia(self.application.SCALE_RESISTOR_100K,
    #                                                 self.application.SCALE_RESISTOR_128K,
    #                                                 self.application.SCALE_RESISTOR_100K)
    #     assert (x["payload"]["status"] == CommonStatus.OK)
    #
    # def test_dynamic_scaling(self):
    #     x = self.application.enable_dynamic_scaling(self.application.SCALE_RESISTOR_100K,
    #                                                 self.application.SCALE_RESISTOR_128K,
    #                                                 self.application.SCALE_RESISTOR_100K)
    #     assert (x["payload"]["status"] == CommonStatus.OK)
    #     x = self.application.disable_dynamic_scaling()
    #     assert (x["payload"]["status"] == CommonStatus.OK)
    #
    # def test_set_discrete_fourier_transformation(self):
    #     x = self.application.set_discrete_fourier_transformation(self.application.DFT_WINDOW_16)
    #     assert (x["payload"]["dft_window"] == self.application.DFT_WINDOW_16)
    #
    # def test_library_configuration(self):
    #     x = self.application.read_library_configuration([0x00])
    #     assert (x["payload"]["size"] == 1)
    #     x = self.application.write_library_configuration([[0x00, 0x1]])
    #     assert (x["payload"]["size"] == 1)
    #
    # def test_device_configuration_block(self):
    #     x = self.application.write_device_configuration_block([[0x0, 2]])
    #     assert (x["payload"]["size"] == 0)
    #     x = self.application.read_device_configuration_block()
    #     assert (x["payload"]["data"] == [['0x0', '0x2']])
    #     x = self.application.delete_device_configuration_block()
    #     assert (x["payload"]["size"] == 0)
    #
    # def test_decimation_factor(self):
    #     x = self.application.set_decimation_factor(2)
    #     assert (x["payload"]["decimation_factor"] == 2)
    #     x = self.application.get_decimation_factor()
    #     assert (type(x["payload"]["decimation_factor"]) == int)
    #
    # def test_get_sensor_status(self):
    #     x = self.application.get_sensor_status()
    #     assert (type(x["payload"]["num_subscribers"]) == int)
    #     assert (type(x["payload"]["num_start_registered"]) == int)
    #
    # def test_get_version(self):
    #     x = self.application.get_version()
    #     assert (type(x["payload"]["major_version"]) == int)
    #
    # def test_stream(self):
    #     common_test.test_stream(self.application)
    #
    # def test_stream_combined(self):
    #     common_test.test_stream_combined(self.application)
    #
    # def test_write_dcb_to_lcfg(self):
    #     x = self.application.write_dcb_to_lcfg()
    #     assert (x["payload"]["status"] == CommonStatus.ERROR)



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

from enum import Enum

from .. import utils


class EDACommand(Enum):
    """
    EDACommand Enum
    """
    LOWEST = [0x40]
    DYNAMIC_SCALE_REQ = [0x42]
    DYNAMIC_SCALE_RES = [0x43]
    SET_DFT_NUM_REQ = [0x46]
    SET_DFT_NUM_RES = [0x47]
    DEBUG_INFO_REQ = [0x48]
    DEBUG_INFO_RES = [0x49]
    RESISTOR_TIA_CAL_REQ = [0x4A]
    RESISTOR_TIA_CAL_RES = [0x4B]
    BASELINE_IMP_SET_REQ = [0x4C]
    BASELINE_IMP_SET_RES = [0x4D]
    BASELINE_IMP_RESET_REQ = [0x4E]
    BASELINE_IMP_RESET_RES = [0x4F]
    LOAD_DCFG_REQ = [0x50]
    LOAD_DCFG_RES = [0x51]
    WRITE_DCFG_REQ = [0x52]
    WRITE_DCFG_RES = [0x53]
    READ_DCFG_REQ = [0x54]
    READ_DCFG_RES = [0x55]
    BASELINE_IMP_GET_REQ = [0x56]
    BASELINE_IMP_GET_RES = [0x57]
    GET_RTIA_TABLE_FDS_REQ = [0x58]
    GET_RTIA_TABLE_FDS_RES = [0x59]
    GET_RTIA_TABLE_RAM_REQ = [0x5A]
    GET_RTIA_TABLE_RAM_RES = [0x5B]
    DELETE_RTIA_TABLE_IN_FDS_REQ = [0x5C]
    DELETE_RTIA_TABLE_IN_FDS_RES = [0x5D]
    CONTROL_AD5940_SLEEP_WAKEUP_REQ = [0x62]
    CONTROL_AD5940_SLEEP_WAKEUP_RES = [0x63]
    SLEEP_WAKEUP_STATUS_REQ = [0x64]
    SLEEP_WAKEUP_STATUS_RES = [0x65]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


class EDAPowerMode(Enum):
    """
    EDAPowerMode Enum
    """
    POWER_INVALID = [0x0]
    POWER_SLEEP = [0x1]
    POWER_WAKEUP = [0x2]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


class EDADFTWindow(Enum):
    """
    EDADFTWindow Enum
    """
    DFT_WINDOW_4 = [0x0]
    DFT_WINDOW_8 = [0x1]
    DFT_WINDOW_16 = [0x2]
    DFT_WINDOW_32 = [0x3]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


class ScaleResistor(Enum):
    """
    ScaleResistor Enum
    """
    DISABLED = [0, 0]
    SCALE_RESISTOR_110 = [1, 0]
    SCALE_RESISTOR_1K = [2, 0]
    SCALE_RESISTOR_2K = [3, 0]
    SCALE_RESISTOR_3K = [4, 0]
    SCALE_RESISTOR_4K = [5, 0]
    SCALE_RESISTOR_6K = [6, 0]
    SCALE_RESISTOR_8K = [7, 0]
    SCALE_RESISTOR_10K = [8, 0]
    SCALE_RESISTOR_12K = [9, 0]
    SCALE_RESISTOR_16K = [10, 0]
    SCALE_RESISTOR_20K = [11, 0]
    SCALE_RESISTOR_24K = [12, 0]
    SCALE_RESISTOR_30K = [13, 0]
    SCALE_RESISTOR_32K = [14, 0]
    SCALE_RESISTOR_40K = [15, 0]
    SCALE_RESISTOR_48K = [16, 0]
    SCALE_RESISTOR_64K = [17, 0]
    SCALE_RESISTOR_85K = [18, 0]
    SCALE_RESISTOR_96K = [19, 0]
    SCALE_RESISTOR_100K = [20, 0]
    SCALE_RESISTOR_120K = [21, 0]
    SCALE_RESISTOR_128K = [22, 0]
    SCALE_RESISTOR_160K = [23, 0]
    SCALE_RESISTOR_196K = [24, 0]
    SCALE_RESISTOR_256K = [25, 0]
    SCALE_RESISTOR_512K = [26, 0]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

import os
import logging
import threading
from queue import Queue

from tqdm import tqdm

from . import utils
from .enums.common_enums import Application, Stream
from .enums.low_touch_enum import CommandType

logger = logging.getLogger(__name__)


class PacketManager:
    """

    PacketManager

    """

    def __init__(self, serial_object, filename=None):
        """
        Initialization of packet Manager

        :param serial_object: Serial port object
        :type serial_object: Serial
        """
        self.queue = Queue()
        self.filename = filename
        self.command_type = None
        self.subscribe_source = {}
        self.lock = threading.Lock()
        self.command_logging = None
        self.log_command_packet = None
        self.source = Application.APP_USB
        self.serial_object = serial_object
        self._kill_threads = threading.Event()
        self.receive_thread = threading.Thread(target=self.receive_packet, name=f"receive_thread")

    def set_usb_source(self):
        self.source = Application.APP_USB

    def set_ble_source(self):
        self.source = Application.APP_BLE

    def enable_command_logging(self, command_type):
        self.command_logging = True
        self.command_type = command_type

    def disable_command_logging(self, command_type):
        self.command_logging = False
        self.command_type = command_type

    def subscribe_command_logger(self, log_command):
        self.log_command_packet = log_command

    def unsubscribe_command_logger(self):
        self.command_type = None
        self.command_logging = None
        self.log_command_packet = None

    def process_queue(self):
        """
        Reads the packet from queue and send to relevant callback function.
        """
        while True:
            try:
                pkt = self.queue.get()
                if self._kill_threads.is_set():
                    break
                pkt = list(pkt)

                logger.info(f"RX :: {':'.join(utils.convert_int_array_to_hex(pkt))}")
                self.send_packet_to_callback(pkt)
            except Exception as e:
                logger.critical(f"Process_queue thread failed : {e}.", exc_info=True)

    def send_packet_to_callback(self, pkt):
        """
        Sends Packet to callback
        """
        try:
            packet_id = pkt[0] + (pkt[1] << 8) + (pkt[8] << 16)
            if pkt[2:4] in [Application.APP_USB.value, Application.APP_BLE.value, Application.FS.value,
                            Application.PM.value]:
                callbacks = list(self.subscribe_source.get(packet_id, []))
                for callback in callbacks:
                    callback(pkt, packet_id)
                if not len(callbacks):
                    logger.warning(f"Received an unexpected packet {':'.join(utils.convert_int_array_to_hex(pkt))} "
                                   f"with packet ID {packet_id}")
        except Exception as e:
            logger.critical(f"RX {':'.join(utils.convert_int_array_to_hex(pkt))} callback error : {e}.", exc_info=True)

    def receive_packet(self):
        """
        Reads packet from serial port and store it in queue.
        """
        while True:
            try:
                if self._kill_threads.is_set():
                    break
                pkt = self.serial_object.read(4)
                if not len(pkt):
                    continue
                if not self._check_valid_source_and_destination_live(pkt):
                    pkt = self._find_valid_source_and_destination_live(pkt, self.serial_object)
                if len(pkt):
                    pkt += self.serial_object.read(4)
                    if len(pkt) >= 8:
                        length = (pkt[4] << 8) + pkt[5]
                        pkt += self.serial_object.read(length - 8)
                        self.queue.put(pkt)
            except Exception as e:
                break

    @staticmethod
    def _find_valid_source_and_destination_live(packet, reader):
        while True:
            new_value = reader.read(1)
            if not len(new_value):
                return []
            packet = packet[1:] + new_value
            if PacketManager._check_valid_source_and_destination_live(packet):
                return packet

    @staticmethod
    def _check_valid_source_and_destination_live(packet):
        destination = list(packet[2:4])
        if destination in [Application.APP_USB.value, Application.APP_BLE.value]:
            return True
        return False

    @staticmethod
    def _find_valid_source_and_destination(packet, reader):
        logger.warning("Finding valid source and destination.")
        while True:
            new_value = reader.read(1)
            if not len(new_value):
                return []
            packet = packet[1:] + new_value
            if PacketManager._check_valid_source_and_destination(packet):
                return packet

    @staticmethod
    def _check_valid_source_and_destination(packet):
        source = list(packet[0:2])
        destination = list(packet[2:4])
        result = (PacketManager._check_stream(source) or PacketManager._check_application(source)) and \
                 (PacketManager._check_stream(destination) or PacketManager._check_application(destination))
        return result

    # noinspection PyBroadException
    @staticmethod
    def _check_stream(packet):
        try:
            source = Stream(list(packet))
            if source == Stream.NULL:
                return False
            return True
        except:
            return False

    # noinspection PyBroadException
    @staticmethod
    def _check_application(packet):
        try:
            source = Application(list(packet))
            if source == Application.NULL:
                return False
            return True
        except:
            return False

    def process_file(self, display_progress, progress_callback):
        """
        Reads packet from file and store it in file queue.
        """
        file_size = os.path.getsize(self.filename)
        progress_bar = None
        current_size = 0
        if display_progress:
            progress_bar = tqdm(total=file_size)
            progress_bar.set_description("Reading file")
        file = open(self.filename, 'rb')
        while True:
            try:
                pkt = file.read(4)
                if not self._check_valid_source_and_destination(pkt):
                    if display_progress:
                        progress_bar.set_description("Recovery Mode")
                    pkt = self._find_valid_source_and_destination(pkt, file)
                    if display_progress:
                        progress_bar.set_description("Reading file")
                    if not len(pkt):
                        break
                pkt += file.read(4)
                if len(pkt):
                    length = (int(pkt[4]) << 8) + int(pkt[5])
                    if length < 8:
                        continue
                    current_size += length
                    if display_progress:
                        progress_bar.update(length)
                    if progress_callback:
                        progress_callback(0, file_size, current_size)
                    pkt += file.read(length - 8)
                    pkt = list(pkt)
                    self.send_packet_to_callback(pkt)
                else:
                    logger.debug(f"EOF reached, exiting loop.")
                    break
            except Exception as e:
                logger.critical(f"process_file method failed : {e}.", exc_info=True)
                break
        if progress_callback:
            progress_callback(0, file_size, file_size)
        file.close()
        if display_progress:
            progress_bar.close()

    def send_packet(self, packet):
        """
        Receives packet from sdk and send it to the serial to port for device to process.
        """
        packet.set_header("source", self.source)
        logger.debug(f"Packet ready : {packet}")
        packet = packet.to_list()
        # logging for LT enable_command_logging
        if self.command_logging is True and self.command_type == CommandType.START:
            self.log_command_packet.add_start_command(packet)
        elif self.command_logging is True and self.command_type == CommandType.STOP:
            self.log_command_packet.add_stop_command(packet)
        else:
            if not self.filename:  # if not convert_log_to_csv
                self.serial_object.write(packet)
        logger.info(f"TX :: {':'.join(utils.convert_int_array_to_hex(packet))}")

    def start_receive_and_process_threads(self):
        """
        start process and receive thread.
        """
        self.receive_thread.setDaemon(True)
        self.receive_thread.start()
        thread = threading.Thread(target=self.process_queue)
        thread.setDaemon(True)
        thread.name = f"process_thread"
        thread.start()

    def subscribe(self, packet_id, callback):
        """
        Subscribe packet id to the callback.
        """
        callbacks = self.subscribe_source.get(packet_id, [])
        callbacks.append(callback)
        self.subscribe_source[packet_id] = callbacks

    def unsubscribe(self, packet_id, application_callback):
        """
        Unsubscribe packet id and removes the callback.
        """
        callbacks = self.subscribe_source.get(packet_id, [])
        for i, callback in enumerate(callbacks):
            if callback == application_callback:
                del callbacks[i]
                break
        self.subscribe_source[packet_id] = callbacks

    def close(self):
        """Closes serial port"""
        self._kill_threads.set()
        # to free up queue.get() lock in process_queue add -1 to queue
        self.queue.put(-1)
        self.serial_object.cancel_read()
        self.serial_object.close()



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

from typing import Dict, List

from .csv_logging import CSVLogger
from .common_stream import CommonStream
from ..core.packets.command_packet import CommandPacket
from ..core.packets.common_packets import VersionPacket
from ..core.enums.pedometer_enums import PedometerCommand
from ..core.packets.stream_data_packets import PedometerDataPacket
from ..core.enums.common_enums import Application, Stream, CommonCommand


class PedometerApplication(CommonStream):
    """
    Pedometer Application class.

    .. code-block:: python3
        :emphasize-lines: 4

        from adi_study_watch import SDK

        sdk = SDK("COM4")
        application = sdk.get_pedometer_application()

    """

    def __init__(self, packet_manager):
        super().__init__(Application.PEDOMETER, Stream.PEDOMETER, packet_manager)

    @staticmethod
    def get_supported_streams() -> List[Stream]:
        """
        List all supported streams.

        :return: Array of stream ID enums.
        :rtype: List[Stream]

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_adpd_application()
            x = application.get_supported_streams()
            print(x)
            # [<Stream.ADPD1: ['0xC2', '0x11']>, ... , <Stream.ADPD12: ['0xC2', '0x1D']>]
        """
        return [Stream.PEDOMETER]

    def get_version(self) -> Dict:
        """
        Returns Pedometer version info.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pedometer_application()
            x = application.get_version()
            print(x["payload"]["major_version"])
            # 3
            print(x["payload"]["minor_version"])
            # 4
            print(x["payload"]["patch_version"])
            # 3
            print(x["payload"]["version_string"])
            # ECG_App
            print(x["payload"]["build_version"])
            # TEST ECG_VERSION STRING
        """
        request_packet = CommandPacket(self._destination, CommonCommand.GET_VERSION_REQ)
        response_packet = VersionPacket(self._destination, CommonCommand.GET_VERSION_RES)
        return self._send_packet(request_packet, response_packet)

    def get_algo_version(self) -> Dict:
        """
        Returns Pedometer version info.

        :return: A response packet as dictionary.
        :rtype: Dict

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pedometer_application()
            x = application.get_algo_version()
            print(x["payload"]["major_version"])
            # 3
            print(x["payload"]["minor_version"])
            # 4
            print(x["payload"]["patch_version"])
            # 3
            print(x["payload"]["version_string"])
            # ECG_App
            print(x["payload"]["build_version"])
            # TEST ECG_VERSION STRING
        """
        request_packet = CommandPacket(self._destination, PedometerCommand.GET_ALGO_VENDOR_VERSION_REQ)
        response_packet = VersionPacket(self._destination, PedometerCommand.GET_ALGO_VENDOR_VERSION_RES)
        return self._send_packet(request_packet, response_packet)

    def _callback_data(self, packet, packet_id, callback_function=None, args=None):
        """
        Process and returns the data back to user's callback function.
        """
        self._callback_data_helper(packet, PedometerDataPacket())

    def enable_csv_logging(self, filename: str, header: List = None) -> None:
        """
        Start logging stream data into CSV.

        :param filename: Name of the CSV file.
        :param header: Header list of the CSV file.
        :return: None

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pedometer_application()
            x = application.enable_csv_logging("ped.csv")
        """
        if header is None:
            header = ["Timestamp", "Steps"]
        self._csv_logger[Stream.PEDOMETER] = CSVLogger(filename, header)

    def disable_csv_logging(self) -> None:
        """
        Stops logging stream data into CSV.

        :return: None

        .. code-block:: python3
            :emphasize-lines: 5

            from adi_study_watch import SDK

            sdk = SDK("COM4")
            application = sdk.get_pedometer_application()
            x = application.disable_csv_logging()
        """
        if self._csv_logger.get(Stream.PEDOMETER):
            self._csv_logger[Stream.PEDOMETER].stop_logging()
        self._csv_logger[Stream.PEDOMETER] = None



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

from unittest import TestCase

from sdk_pure_python.adi_study_watch import SDK
from sdk_pure_python.adi_study_watch.core.enums.common_enums import CommonStatus


class TestAD7156Application(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.application = SDK("COM4").get_ad7156_application()

    def test_load_configuration(self):
        x = self.application.load_configuration()
        assert (x["payload"]["status"] == CommonStatus.OK)

    def test_device_configuration_block(self):
        x = self.application.write_device_configuration_block([[0x10, 2], [0x11, 0x1]])
        print(x)
        x = self.application.read_device_configuration_block()
        assert (x["payload"]["data"] == [['0x10', '0x2'], ['0x11', '0x1']])
        x = self.application.delete_device_configuration_block()
        print(x)

    def test_register(self):
        x = self.application.write_register([[0xa, 0x5], [0xb, 0x2]])
        assert (x["payload"]["size"] == 2)
        assert (x["payload"]["data"] == [['0xA', '0x5'], ['0xB', '0x2']])
        x = self.application.read_register([0xa, 0xb])
        assert (x["payload"]["size"] == 2)
        assert (x["payload"]["data"] == [['0xA', '0x5'], ['0xB', '0x2']])



# ******************************************************************************
# Copyright (c) 2019 Analog Devices, Inc.  All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# - Redistributions of source code must retain the above copyright notice, this
#  list of conditions and the following disclaimer.
# - Redistributions in binary form must reproduce the above copyright notice,
#  this list of conditions and the following disclaimer in the documentation
#  and/or other materials provided with the distribution.
# - Modified versions of the software must be conspicuously marked as such.
# - This software is licensed solely and exclusively for use with
#  processors/products manufactured by or for Analog Devices, Inc.
# - This software may not be combined or merged with other code in any manner
#  that would cause the software to become subject to terms and conditions
#  which differ from those listed here.
# - Neither the name of Analog Devices, Inc. nor the names of its contributors
#  may be used to endorse or promote products derived from this software
#  without specific prior written permission.
# - The use of this software may or may not infringe the patent rights of one
#  or more patent holders.  This license does not release you from the
#  requirement that you obtain separate licenses from these patent holders to
#  use this software.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES, INC. AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
# NONINFRINGEMENT, TITLE, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL ANALOG DEVICES, INC. OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, PUNITIVE OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, DAMAGES ARISING OUT OF
# CLAIMS OF INTELLECTUAL PROPERTY RIGHTS INFRINGEMENT; PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# ******************************************************************************

from enum import Enum, unique

from .adp5360_enums import ADP5360Command
from .lt_mode4_enums import LTMode4Command, LTMode4Status
from .. import utils
from .sqi_enum import SQICommand
from .bia_enums import BIACommand
from .ecg_enums import ECGCommand
from .eda_enums import EDACommand
from .ppg_enums import PPGCommand
from .adxl_enums import ADXLCommand
from .adpd_enums import ADPDCommand
from .low_touch_enum import LTCommand
from .ad7156_enums import AD7156Command
from .fs_enums import FSCommand, FSStatus
from .display_enums import DisplayCommand
from .dcb_enums import DCBCommand, DCBStatus
from .pedometer_enums import PedometerCommand
from .user0_enums import User0Command, User0Status
from .pm_enums import PMCommand, PMStatus, UICRCustomerRegisterAccessStatus
from .session_manager_enums import SessionManagerStatus, SessionManagerCommand
from .low_touch_enum import LTStatus

def get_command(command, source):
    """
    Helper method for decoding commands.
    """
    if command[0] < CommonCommand.HIGHEST.value[0]:
        return CommonCommand
    elif command in [DCBCommand.READ_CONFIG_RES.value, DCBCommand.WRITE_CONFIG_RES.value,
                     DCBCommand.ERASE_CONFIG_RES.value, DCBCommand.QUERY_STATUS_RES.value]:
        return DCBCommand
    elif source == Application.PM:
        return PMCommand
    elif source == Application.ADPD:
        return ADPDCommand
    elif source == Application.ADXL:
        return ADXLCommand
    elif source == Application.ADP5360:
        return ADP5360Command
    elif source == Application.FS or source == Stream.FS:
        return FSCommand
    elif source == Application.PPG:
        return PPGCommand
    elif source == Application.SQI:
        return SQICommand
    elif source == Application.DISPLAY:
        return DisplayCommand
    elif source == Application.AD7156:
        return AD7156Command
    elif source == Application.BIA:
        return BIACommand
    elif source == Application.EDA:
        return EDACommand
    elif source == Application.LT_APP:
        return LTCommand
    elif source == Application.ECG:
        return ECGCommand
    elif source == Application.PEDOMETER:
        return PedometerCommand
    elif source == Application.USER0_APP:
        return User0Command
    elif source == Application.SESSION_MANAGER_APP:
        return SessionManagerCommand
    elif source == Application.LT_MODE4_APP:
        return LTMode4Command
    else:
        return CommonCommand


def get_status(status, source, command):
    """
    Helper method for decoding status.
    """
    if command[0] == CommonCommand.ALARM_NOTIFICATION.value[0]:
        return AlarmStatus
    elif command in [PMCommand.WRITE_UICR_CUSTOMER_REG_RES.value, PMCommand.READ_UICR_CUSTOMER_REG_RES.value]:
        return UICRCustomerRegisterAccessStatus
    elif status[0] < CommonStatus.HIGHEST.value[0]:
        return CommonStatus
    elif command in [DCBCommand.READ_CONFIG_RES.value, DCBCommand.WRITE_CONFIG_RES.value,
                     DCBCommand.ERASE_CONFIG_RES.value]:
        return DCBStatus
    elif source == Application.PM:
        return PMStatus
    elif source == Application.USER0_APP:
        return User0Status
    elif source == Application.LT_MODE4_APP:
        return LTMode4Status
    elif source == Application.FS or source == Stream.FS:
        return FSStatus
    elif source == Application.SESSION_MANAGER_APP:
        return SessionManagerStatus
    elif source == Application.LT_APP:
        return LTStatus
    else:
        return CommonStatus


@unique
class CommonCommand(Enum):
    """
    CommonCommand Enum
    """
    NO_RESPONSE = [-1]
    GET_VERSION_REQ = [0x0]
    GET_VERSION_RES = [0x1]
    START_SENSOR_REQ = [0x4]
    START_SENSOR_RES = [0x5]
    STOP_SENSOR_REQ = [0x6]
    STOP_SENSOR_RES = [0x7]
    SUBSCRIBE_STREAM_REQ = [0xC]
    SUBSCRIBE_STREAM_RES = [0xD]
    UNSUBSCRIBE_STREAM_REQ = [0xE]
    UNSUBSCRIBE_STREAM_RES = [0xF]
    GET_SENSOR_STATUS_REQ = [0x10]
    GET_SENSOR_STATUS_RES = [0x11]
    GET_LCFG_REQ = [0x12]
    GET_LCFG_RES = [0x13]
    SET_LCFG_REQ = [0x14]
    SET_LCFG_RES = [0x15]
    READ_LCFG_REQ = [0x16]
    READ_LCFG_RES = [0x17]
    WRITE_LCFG_REQ = [0x18]
    WRITE_LCFG_RES = [0x19]
    PING_REQ = [0x1A]
    PING_RES = [0x1B]
    ALARM_NOTIFICATION = [0x1C]
    REGISTER_READ_REQ = [0x21]
    REGISTER_READ_RES = [0x22]
    REGISTER_WRITE_REQ = [0x23]
    REGISTER_WRITE_RES = [0x24]
    GET_DCFG_REQ = [0x25]
    GET_DCFG_RES = [0x26]
    STREAM_DATA = [0x28]
    GET_STREAM_DEC_FACTOR_REQ = [0x29]
    GET_STREAM_DEC_FACTOR_RES = [0x2A]
    SET_STREAM_DEC_FACTOR_REQ = [0x2B]
    SET_STREAM_DEC_FACTOR_RES = [0x2C]
    REGISTER_READ_32_REQ = [0x2D]
    REGISTER_READ_32_RES = [0x2E]
    REGISTER_WRITE_32_REQ = [0x2F]
    REGISTER_WRITE_32_RES = [0x30]
    HIGHEST = [0x40]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


@unique
class CommonStatus(Enum):
    """
    CommonStatus Enum
    """
    NO_RESPONSE = [-1]
    OK = [0x0]
    ERROR = [0x1]
    STREAM_STARTED = [0x2]
    STREAM_STOPPED = [0x3]
    STREAM_IN_PROGRESS = [0x4]
    STREAM_DEACTIVATED = [0x5]
    STREAM_COUNT_DECREMENT = [0x6]
    STREAM_NOT_STARTED = [0x7]
    STREAM_NOT_STOPPED = [0x8]
    SUBSCRIBER_ADDED = [0x9]
    SUBSCRIBER_REMOVED = [0xA]
    SUBSCRIBER_COUNT_DECREMENT = [0xB]
    STREAM_NOT_STARTED_BATTERY_LOW = [0xC]
    SUBSCRIBER_NOT_ADDED_BATTERY_LOW = [0xD]
    HIGHEST = [0x20]
    NEW_STREAM_STATUS = [0x43]  # some stream such as EDA has 0x43 as status instead of 0x0

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


@unique
class Stream(Enum):
    """
    Stream Enum
    """
    NULL = [0x00, 0x00]
    ADPD1 = [0xc2, 0x11]
    ADPD2 = [0xc2, 0x12]
    ADPD3 = [0xc2, 0x13]
    ADPD4 = [0xc2, 0x14]
    ADPD5 = [0xc2, 0x15]
    ADPD6 = [0xc2, 0x16]
    ADPD7 = [0xc2, 0x17]
    ADPD8 = [0xc2, 0x18]
    ADPD9 = [0xc2, 0x19]
    ADPD10 = [0xc2, 0x1a]
    ADPD11 = [0xc2, 0x1b]
    ADPD12 = [0xc2, 0x1c]
    ADXL = [0xc2, 0x02]
    BIA = [0xC4, 0x07]
    BCM = [0xC8, 0x14]
    ECG = [0xc4, 0x01]
    EDA = [0xc4, 0x02]
    FS = [0xC6, 0x01]
    PEDOMETER = [0xc4, 0x04]
    PPG = [0xC4, 0x00]
    # TEMPERATURE = [0xc4, 0x06]
    SYNC_PPG = [0xC4, 0x05]
    SQI = [0xC8, 0x0D]
    BATTERY = [0xC6, 0x91]
    DYNAMIC_AGC_STREAM = [0xC6, 0xB0]
    STATIC_AGC_STREAM = [0xC6, 0xB1]
    HRV = [0xC6, 0xC0]
    AD7156 = [0xC8, 0x15]
    TEMPERATURE1 = [0xC8, 0x16]
    TEMPERATURE2 = [0xC8, 0x17]
    TEMPERATURE3 = [0xC8, 0x18]  # C
    TEMPERATURE4 = [0xC4, 0x06]  # D
    TEMPERATURE5 = [0xC8, 0x1A]
    TEMPERATURE6 = [0xC8, 0x1B]
    TEMPERATURE7 = [0xC8, 0x1C]
    TEMPERATURE8 = [0xC8, 0x1D]
    TEMPERATURE9 = [0xC8, 0x1E]
    TEMPERATURE10 = [0xC8, 0x1F]  # J
    TEMPERATURE11 = [0xC8, 0x20]  # K
    TEMPERATURE12 = [0xC8, 0x21]  # L

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


@unique
class Application(Enum):
    """
    Application Enum.
    """
    POST_OFFICE = [0xC0, 0x0]
    EXTERNAL = [0xC0, 0x01]
    ADXL = [0xc1, 0x02]
    ADP5360 = [0xc1, 0x06]
    ADPD = [0xc1, 0x10]
    PPG = [0xC3, 0x00]
    ECG = [0xc3, 0x01]
    EDA = [0xc3, 0x02]
    PEDOMETER = [0xc3, 0x04]
    TEMPERATURE = [0xc3, 0x06]
    BIA = [0xC3, 0x07]
    FS = [0xC5, 0x01]
    PS = [0xc5, 0x80]
    PM = [0xc5, 0x00]
    DISPLAY = [0xC5, 0x03]
    APP_ANDROID = [0xC7, 0x0]
    APP_IOS = [0xC7, 0x1]
    APP_VS = [0xC7, 0x2]
    APP_WT = [0xC7, 0x3]
    APP_NFE = [0xC7, 0x4]
    APP_USB = [0xc7, 0x05]
    SQI = [0xC8, 0x0C]
    AD7156 = [0xC8, 0x0B]
    APP_BLE = [0xC8, 0x08]
    LT_APP = [0xC8, 0x0A]
    USER0_APP = [0xC8, 0x0F]
    LT_MODE4_APP = [0xC8, 0x24]
    SESSION_MANAGER_APP = [0xC8, 0x25]
    NULL = [0x0, 0x0]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))


@unique
class AlarmStatus(Enum):
    BATTERY_LEVEL_LOW = [0x0]
    BATTERY_LEVEL_CRITICAL = [0x1]
    BATTERY_LEVEL_FULL = [0x2]
    NAND_LT_CONFIG_LOG_ENABLED = [0x3]
    ENABLE_NAND_LT_CONFIG_LOG_FAILED = [0x4]
    NAND_LT_CONFIG_LOG_DISABLED = [0x5]
    DISABLE_NAND_LT_CONFIG_LOG_FAILED = [0x6]
    DCB_LT_CONFIG_LOG_ENABLED = [0x7]
    ENABLE_DCB_LT_CONFIG_LOG_FAILED = [0x8]
    DCB_LT_CONFIG_LOG_DISABLED = [0x9]
    DISABLE_DCB_LT_CONFIG_LOG_FAILED = [0xA]
    LOW_TOUCH_LOGGING_ALREADY_STARTED = [0xB]
    LT_CONFIG_FILE_NOT_FOUND = [0xC]
    LT_CONFIG_FILE_READ_ERR = [0xD]
    LOW_TOUCH_MEMORY_FULL_ERR = [0xE]
    LOW_TOUCH_MAX_FILE_ERR = [0xF]
    FS_MEMORY_FULL = [0x10]
    FS_LOG_STOP_BATTERY_LOW = [0x11]
    FS_LOG_DOWNLOAD_STOP_BATTERY_CRITICAL = [0x12]
    FS_PWR_STATE_SHUTDOWN = [0x13]
    BLE_PKT_MISS = [0x14]
    FS_FILE_CLOSE_ERROR = [0x15]
    FS_FILE_WRITE_ERROR = [0x16]
    LT_LCFG_DCB_READ_ERROR = [0x17]
    USER0_LCFG_DCB_READ_ERROR = [0x18]
    ADPD_CONT_SATURATION_DETECTED = [0x19]
    ECG_LEADS_OFF_DETECTED = [0x1A]
    ECG_LEADS_ON_DETECTED = [0x1B]
    CHARGER_OFF_DETECTED = [0x1C]
    CHARGER_ON_DETECTED = [0x1D]
    NO_ALARM  = [0x1E]
    HIGHEST = [0x1F]

    def __repr__(self):
        return "<%s.%s: %r>" % (self.__class__.__name__, self._name_, utils.convert_int_array_to_hex(self._value_))
